package com.laguerre.kan;

import java.util.ArrayList;
import java.util.Arrays;

import com.laguerre.kan.SceneConductor.ISurfaceChangedListener;

import android.opengl.GLES20;
import android.opengl.Matrix;

public final class Camera extends Component implements ISurfaceChangedListener {
	/**
	 * Contains values that are invariant for a {@link Camera} that renders
	 * objects, such as the view matrix, the projection matrix, the concatenation
	 * of the view and the projection matrix, or the eye/camera position.
	 */
	public static final class RenderingValues {
		private final int[] viewSize = new int[2];
		private float[] nearAndFarClipPlanes = new float[2];
		private final Vector3 eyePosition = Vector3.getNewZeroVec();
		private final float[] viewMatrix = new float[16];
		private final float[] invViewMatrix = new float[16];
		private final float[] projMatrix = new float[16];
		private final float[] invProjMatrix = new float[16];
		private final float[] viewProjMatrix = new float[16];
		private final float[] invViewProjMatrix = new float[16];

		RenderingValues() {
			Arrays.fill(this.viewSize, 100);
			Arrays.fill(this.nearAndFarClipPlanes, 0f);
			Matrix.setIdentityM(this.viewMatrix, 0);
			Matrix.setIdentityM(this.invViewMatrix, 0);
			Matrix.setIdentityM(this.projMatrix, 0);
			Matrix.setIdentityM(this.invProjMatrix, 0);
			Matrix.setIdentityM(this.viewProjMatrix, 0);
			Matrix.setIdentityM(this.invViewProjMatrix, 0);
		}

		/**
		 * Sets the size of the view in pixels.
		 * @param width
		 * @param height
		 */
		public void setViewSize(int width, int height) {
			this.viewSize[0] = width;
			this.viewSize[1] = height;
		}

		/**
		 * Sets the size of the view in pixels.
		 * @param viewSize where viewSize[0] contains width and viewSize[1]
		 *   contains height.
		 */
		public void setViewSize(int[] viewSize) {
			System.arraycopy(viewSize, 0, this.viewSize, 0, 2);
		}

		/**
		 * Sets width of the view in pixels.
		 * @param width
		 */
		public void setViewSizeWidth(int width) {
			this.viewSize[0] = width;
		}

		/**
		 * Sets height of the view in pixels.
		 * @param width
		 */
		public void setViewSizeHeight(int height) {
			this.viewSize[1] = height;
		}

		/**
		 * Gets a direct reference to the inner <code>int[]</code> that
		 * contains view size in pixels in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the view size.
		 */
		public int[] getSharedViewSize() {
			return this.viewSize;
		}

		/**
		 * Gets the width of this view.
		 * @return the width.
		 */
		public int getWidth() {
			return this.viewSize[0];
		}

		/**
		 * Gets the height of this view.
		 * @return the height.
		 */
		public int getHeight() {
			return this.viewSize[1];
		}

		/**
		 * Sets the near and far clip planes.
		 * @param near the near clip plane.
		 * @param far the far clip plane.
		 */
		public void setNearAndFarClipPlanes(float near, float far) {
			this.nearAndFarClipPlanes[0] = near;
			this.nearAndFarClipPlanes[1] = far;
		}

		/**
		 * Sets the near clip plane.
		 * @param near the near clip plane.
		 */
		public void setNearClipPlane(float near) {
			this.nearAndFarClipPlanes[0] = near;
		}

		/**
		 * Sets the far clip plane.
		 * @param far the far clip plane.
		 */
		public void setFarClipPlane(float far) {
			this.nearAndFarClipPlanes[1] = far;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains near and far clip planes in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the near and far clip planes.
		 */
		public float[] getSharedNearAndFarClipPlanes() {
			return this.nearAndFarClipPlanes;
		}

		/**
		 * Gets the near clip plane.
		 * @return the near clip plane.
		 */
		public float getNearClipPlane() {
			return this.nearAndFarClipPlanes[0];
		}

		/**
		 * Gets the far clip plane.
		 * @return the far clip plane.
		 */
		public float getFarClipPlane() {
			return this.nearAndFarClipPlanes[1];
		}

		/**
		 * Sets the eye/camera position in world space.
		 * @param position
		 */
		public void setEyePosition(Vector3 position) {
			this.eyePosition.setValues(position);
		}

		/**
		 * Gets a direct reference to the inner <code>Vector3</code> that
		 * contains eye position in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the eye position.
		 */
		public Vector3 getSharedEyePosition() {
			return this.eyePosition;
		}

		/**
		 * Sets the view matrix and computes its inverse automatically.
		 * @param viewMatrix
		 */
		public void setViewMatrix(float[] viewMatrix) {
			setViewMatrix(viewMatrix, null);
		}

		/**
		 * Sets the view matrix and its inverse. If the inverse equals to
		 * <code>null</code>, it will be computed automatically.
		 * @param viewMatrix the view matrix.
		 * @param invViewMatrix the inverse of the view matrix or <code>null</code>.
		 */
		public void setViewMatrix(float[] viewMatrix, float[] invViewMatrix) {
			setMatrixes(viewMatrix, this.viewMatrix, invViewMatrix, this.invViewMatrix);
		}

		/**
		 * Sets the view-projection matrix and computes its inverse automatically.
		 * @param projViewMatrix
		 */
		public void setViewProjMatrix(float[] projViewMatrix) {
			setViewProjMatrix(projViewMatrix, null);
		}

		/**
		 * Sets the view-projection matrix and its inverse. If the inverse equals
		 * to <code>null</code>, it will be computed automatically.
		 * @param viewProjMatrix the view-projection matrix.
		 * @param invViewProjMatrix the inverse of the view-projection matrix or <code>null</code>.
		 */
		public void setViewProjMatrix(float[] viewProjMatrix, float[] invViewProjMatrix) {
			setMatrixes(viewProjMatrix, this.viewProjMatrix, invViewProjMatrix, this.invViewProjMatrix);
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the view matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the view matrix.
		 */
		public float[] getSharedViewMatrix() {
			return this.viewMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the inverse view matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the inverse view matrix.
		 */
		public float[] getSharedInvViewMatrix() {
			return this.invViewMatrix;
		}

		/**
		 * Sets the projection matrix and computes its inverse automatically.
		 * @param projMatrix
		 */
		public void setProjMatrix(float[] projMatrix) {
			setProjMatrix(projMatrix, null);
		}

		/**
		 * Sets the projection matrix and its inverse. If the inverse equals
		 * to <code>null</code>, it will be computed automatically.
		 * @param projMatrix the projection matrix.
		 * @param invProjMatrix the inverse of the projection matrix or <code>null</code>.
		 */
		public void setProjMatrix(float[] projMatrix, float[] invProjMatrix) {
			setMatrixes(projMatrix, this.projMatrix, invProjMatrix, this.invProjMatrix);
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the projection matrix.
		 */
		public float[] getSharedProjMatrix() {
			return this.projMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the inverse projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the inverse projection matrix.
		 */
		public float[] getSharedInvProjMatrix() {
			return this.invProjMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the view-projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the view-projection matrix.
		 */
		public float[] getSharedViewProjMatrix() {
			return this.viewProjMatrix;
		}

		/**
		 * Gets a direct reference to the inner <code>float[]</code> that
		 * contains the inverse view-projection matrix in order to avoid memory allocations.
		 * Please, do <b>not</b> update its components by this way.
		 * @return A reference to the inverse view-projection matrix.
		 */
		public float[] getSharedInvViewProjMatrix() {
			return this.invViewProjMatrix;
		}

		/**
		 * Fills a <code>float[]</code> array that represents a matrix and also
		 * set its inverse: to the proposed value or automatically if the value
		 * is <code>null</code>.
		 * @param src the matrix to copy
		 * @param fieldDst the reference to the array that will contain the copy of <code>src</code>.
		 * @param invSrc the inverse of the matrix to copy, or <code>null</code> to compute it automatically.
		 * @param fieldInvDst the reference to the array that will contain the copy of the inverse matrix.
		 */
		private void setMatrixes(float[] src, float[] fieldDst, float[] invSrc, float[] fieldInvDst) {
			System.arraycopy(src, 0, fieldDst, 0, src.length);
			if (invSrc != null) {
				System.arraycopy(invSrc, 0, fieldInvDst, 0, src.length);
			} else {
				Matrix.invertM(fieldInvDst, 0, src, 0);
			}
		}
	}

	private static Camera currentCamera = null;
	public static Camera getCurrent() {
		return currentCamera;
	}

	private boolean enabled;
	private RenderTexture renderTexture;
	private final ImageEffects postEffects = new ImageEffects();
	private int fbHandle;
	private int depthFbHandle;
	/**
	 * The default value of the culling mask is <code>1&lt;&lt;0</code>.
	 */
	private int cullingMask = 1 << 0;

	private final Color backgroundColor = Color.GREY;
	private float fovY = 45f;
	private float nearClipPlane = -0.1f;
	private float farClipPlane = -20f;
	private float aspect = 1.0f;
	private boolean orthographic = false;
	private Shader replacementShader = null;
	private String replacementTag = null;

	private final float[] viewMatrix = new float[16];
	private final float[] projMatrix = new float[16];
	private final float[] viewProjMatrix = new float[16];

	private final RenderingValues renderingPrms = new RenderingValues();

	Camera() {
		this.enabled = true;
		this.renderTexture = null;
		this.fbHandle = -1;
		this.depthFbHandle = -1;

		setBackgroundColor(0.5f, 0.5f, 0.5f, 1.0f);
		// Defaults to rendering to texture.
		this.postEffects.setRenderingMode(ImageEffects.FinalDestination.TEXTURE);
	}

	public void setEnabled(boolean enabling) {
		this.enabled = enabling;
	}

	public boolean isEnabled() {
		return this.enabled;
	}

	public void setRenderTexture(RenderTexture renderTexture) {
		if (this.renderTexture != renderTexture) {
			this.renderTexture = renderTexture;
			setWidthAndHeight();
			// Init methods may reset the rendering mode, so backup here for future restore.
			final ImageEffects.FinalDestination renderingMode = this.postEffects.getRenderingMode();
			setupFBO();
			if (!this.postEffects.isInitialized()) {
				this.postEffects.init();
			}
			this.postEffects.setRenderTexture(renderTexture);
			// Restore rendering mode.
			this.postEffects.setRenderingMode(renderingMode);
		}
	}

	private void setupFBO() {
		// Deletes the previous FBO, if any.
		freeResources();

		final int[] names = { -1 };
		// FB
		GLES20.glGenFramebuffers(1, names, 0);
		this.fbHandle = names[0];
		// depth
		GLES20.glGenRenderbuffers(1, names, 0);
		this.depthFbHandle = names[0];
		GLES20.glBindRenderbuffer(GLES20.GL_RENDERBUFFER, this.depthFbHandle);
		GLES20.glRenderbufferStorage(
				GLES20.GL_RENDERBUFFER, GLES20.GL_DEPTH_COMPONENT16,
				this.renderTexture.getWidth(), this.renderTexture.getHeight());
		GLES20.glBindRenderbuffer(GLES20.GL_RENDERBUFFER, 0);
	}

	/**
	 * Free resources used by this camera: the frame buffer and render buffers
	 * are deleted.
	 */
	public void freeResources() {
		final int[] rendBuff = new int[] { this.fbHandle };
		if (this.fbHandle > 0) {
			GLES20.glDeleteFramebuffers(1, rendBuff, 0);
			this.fbHandle = -1;
		}
		if (this.depthFbHandle > 0) {
			rendBuff[0] = this.depthFbHandle;
			GLES20.glDeleteRenderbuffers(1, rendBuff, 0);
			this.depthFbHandle = 1;
		}
		// Same thing for post effects.
		this.postEffects.freeResources();
	}

	public RenderTexture getSharedRenderTexture() {
		return this.renderTexture;
	}

	public void setAsRenderingOnScreen() {
		if (currentCamera != null) {
			currentCamera.postEffects.setRenderingMode(ImageEffects.FinalDestination.TEXTURE);
			SceneConductor.getInstance().removeSurfaceChangedListener(currentCamera);
			currentCamera.setWidthAndHeight();
		}
		currentCamera = this;
		this.postEffects.setRenderingMode(ImageEffects.FinalDestination.SCREEN);
		SceneConductor.getInstance().addSurfaceChangedListener(this);
		setWidthAndHeight();
	}

	public boolean isRenderingOnScreen() {
		return this == currentCamera;
	}

	/**
	 * Sets the culling mask. The default value of the culling mask is
	 * <code>1&lt;&lt;0</code>.
	 * @param mask
	 * @see getCullingMask
	 * @see isMaskCulled
	 */
	public void setCullingMask(int mask) {
		this.cullingMask = mask;
	}

	/**
	 * Gets the culling mask.
	 * @return the value of the culling mask.
	 * @see setCullingMask
	 * @see isMaskCulled
	 */
	public int getCullingMask() {
		return this.cullingMask;
	}

	/**
	 * If the bitwise <code>AND</code> operation between this mask and a
	 * {@link SceneObject} layer returns <code>0</code>, then the
	 * {@link SceneObject} will not be rendered.
	 * @see setCullingMask
	 * @see getCullingMask
	 */
	public boolean isMaskCulled(SceneObject sob) {
		return (sob.getLayer() & this.cullingMask) == 0;
	}

	@Override
	public void onSurfaceChanged(int width, int height) {
		if (isRenderingOnScreen()) {
			setWidthAndHeight();
		}
	}

	/**
	 * Sets <code>width</code> and <code>height</code> values that will be
	 * returned by <code>getWidth</code> and <code>getHeight</code>. The method
	 * will also call <code>updateCameraProperties</code>.
	 *
	 * @see getWidth
	 * @see getHeight
	 * @see updateCameraProperties
	 */
	private void setWidthAndHeight() {
		if (isRenderingOnScreen()) {
			this.renderingPrms.setViewSize(
					SurfaceInfo.getInstance().getSurfaceWidth(),
					SurfaceInfo.getInstance().getSurfaceHeight());
		} else {
			this.renderingPrms.setViewSize(
					renderTexture.getWidth(),
					renderTexture.getHeight());
		}
		updateCameraProperties();
	}

	/**
	 * Gets the width of the view of this {@link Camera}.
	 * @return the width.
	 */
	public int getWidth() {
		return this.renderingPrms.getWidth();
	}

	/**
	 * Gets the height of the view of this {@link Camera}.
	 * @return the height.
	 */
	public int getHeight() {
		return this.renderingPrms.getHeight();
	}

	/**
	 * Sets the near and far clip plane positions on <code>Z</code> from the
	 * origin of this {@link Camera}. Since the camera is pointing to negative
	 * values along <code>Z</code>, values should be very likely negative.
	 * @param near the near position from the origin of this {@link Camera}.
	 * @param far the far position from the origin of this {@link Camera}.
	 */
	public void setNearAndFarClipPlanes(float near, float far) {
		this.nearClipPlane = near;
		this.farClipPlane = far;
		updateCameraProperties();
	}

	/**
	 * Gets the position of the near clip plane on <code>Z</code>.
	 * @return the position.
	 */
	public float getNearClipPlane() {
		return this.nearClipPlane;
	}

	/**
	 * Gets the position of the far clip plane on <code>Z</code>.
	 * @return the position.
	 */
	public float getFarClipPlane() {
		return this.farClipPlane;
	}

	/**
	 * Makes this {@link Camera} orthographic or perspective.
	 * @param ortographic makes this {@link Camera} orthographic if <code>true</code>,
	 *   perspective otherwise.
	 */
	public void setAsOrthographic(boolean ortographic) {
		this.orthographic = ortographic;
		updateCameraProperties();
	}

	/**
	 * Tells whether this {@link Camera} is orthographic or perspective.
	 * @return <code>true</code> if this {@link Camera} is orthographic,
	 *   <code>false</code> otherwise.
	 */
	public boolean isOrthographic() {
		return this.orthographic;
	}

	/**
	 * Gets the current projection matrix.
	 * @param result is a <code>float</code> array of <code>16</code> elements
	 *   that will contain the projection matrix.
	 * @return a reference to <code>result</code>.
	 */
	public float[] getProjMatrix(float[] result) {
		System.arraycopy(this.projMatrix, 0, result, 0, 16);
		return result;
	}

	public float[] getSharedProjMatrix() {
		return this.projMatrix;
	}

	protected void updateCameraProperties() {
		final int width = getWidth();
		final int height = getHeight();

		this.aspect = ((float) width) / height;

		if (this.orthographic) {
			Matrix.orthoM(this.projMatrix, 0,
					-this.aspect, this.aspect, -1f, 1f,
					Math.abs(this.nearClipPlane), Math.abs(this.farClipPlane));
		} else {
//			// Code with Matrix.frustumM.
//			final float halfHeightNear =
//					(float) (Math.abs(this.nearClipPlane) * Math.tan(Math.toRadians(this.fovY / 2.0)));
//			final float halfWidthNear = halfHeightNear * this.aspect;
//			Matrix.frustumM(this.projMatrix, 0,
//					-halfWidthNear, halfWidthNear, -halfHeightNear, halfHeightNear,
//					Math.abs(this.nearClipPlane), Math.abs(this.farClipPlane));

			// Code with perspectiveM (a bit faster that frustumM).
			perspectiveM(this.projMatrix, 0,
					this.fovY, this.aspect,
					Math.abs(this.nearClipPlane), Math.abs(this.farClipPlane));
		}

		computeFrustumPlaneEquations();

		// Sets the projection matrix.
		this.renderingPrms.setProjMatrix(this.projMatrix);
	}

	/**
	 * Sets the color used to fill the background before any drawing starts.
	 * @param color the color to be used.
	 */
	public void setBackgroundColor(Color color) {
		this.backgroundColor.setValues(color);
	}

	/**
	 * Sets the color used to fill the background before any drawing starts.
	 * @param r the red component in [0.0f, 1.0f].
	 * @param g the green component in [0.0f, 1.0f].
	 * @param b the blue component in [0.0f, 1.0f].
	 * @param a the alpha component in [0.0f, 1.0f].
	 */
	public void setBackgroundColor(float r, float g, float b, float a) {
		this.backgroundColor.setValues(r, g, b, a);
	}

	/**
	 * Gets a direct reference to the object that owns the background color of
	 * this {@link Camera} for seeking good performance. Do not attempt to modify
	 * the {@link Color} by this mean.
	 * @return a reference to the background color object.
	 */
	public Color getSharedBackgroundColor() {
		return this.backgroundColor;
	}

	/**
	 * Gets the background color of this {@link Camera}.
	 * @param result will contain the values.
	 * @return a reference to <code>result</code>.
	 */
	public Color getBackgroundColor(Color result) {
		result.setValues(this.backgroundColor);
		return result;
	}

	/**
	 * Gets a copy of the background color of this {@link Camera}.
	 * @return new allocated object that contains the {@link Color}.
	 */
	public Color getBackgroundColor() {
		return new Color(this.backgroundColor);
	}

	/**
	 * Sets a {@link Shader} that will replace all the shaders of {@link SceneObject}
	 * objects to be rendered by this {@link Camera} for those that use the same
	 * tag as <code>replacementTag</code>.
	 * @param shader for replacement.
	 * @param replacementTag the tag that {@link Shader} objects must return by
	 *  <code>getTagValue("RenderType")</code> to see them replaced for this
	 *  {@link Camera}.
	 *  If the value is <code>null</code> or an empty {@link String} then every
	 *  {@link Shader} will be replaced, otherwise it is not rendered.
	 *
	 *  @see resetReplacementShader
	 *  @see Shader.CommonShaderTagTypes
	 *  @see Shader.CommonShaderTags
	 */
	public void setReplacementShader(Shader shader, String replacementTag) {
		this.replacementShader = shader;
		this.replacementTag = replacementTag;
	}

	/**
	 * Gets a direct reference to the current replacement {@link Shader}. The
	 * value will be <code>null</code> if none is used.
	 * @return the current replacement {@link Shader}.
	 */
	public Shader getSharedReplacementShader() {
		return this.replacementShader;
	}

	/**
	 * Gets the tag that is used to check if a replacement shader should
	 * replace a {@link Shader}. The value will be <code>null</code>
	 * if no replacement {@link Shader} is set.
	 * @return the replacement tag.
	 */
	public String getSharedReplacementTag() {
		return this.replacementTag;
	}

	/**
	 * Removes the use of a replacement {@link Shader} (and removes the corresponding
	 * replacement tag).
	 */
	public void resetReplacementShader() {
		this.replacementShader = null;
		this.replacementTag = null;
	}

	/**
	 * Calls <code>renderToTexture</code> and <code>renderOnePostEffectPass</code>
	 * to render one frame and then its post-effects (if any).
	 */
	public void render() {
		renderToTexture();
		setToFirstPostEffectPass();
		while (postEffectStillNeedsPass()) {
			getSharedSOb().onBeforeRenderingImageEffectsPass(getCurrentPostEffectPass());
			renderOnePostEffectPass();
			Utils.checkGlError("Detection of OpenGL error after post-effect");
		}
	}

	/**
	 * Visits all the current {@link SceneObject} objects to find which ones are
	 * visible by this {@link Camera}.
	 */
	private static final class SceneObjectVisitorForVisible implements SceneConductor.IRenderQueueVisitor {
		private final Camera camera;
		private ArrayList<SceneObject> visibleObjects;

		public SceneObjectVisitorForVisible(Camera camera) {
			this.camera = camera;
		}

		public void setResultArray(ArrayList<SceneObject> visibleObjects) {
			this.visibleObjects = visibleObjects;
		}

		@Override
		public void visit(SceneObject sob, int queue) {
			if (sob.isEnabled() && !this.camera.isMaskCulled(sob) &&
					sob.getSharedRenderer() != null) {
				final Renderer renderer = sob.getSharedRenderer();
				renderer.updateWorldBounds();
				if (this.camera.isVisible(renderer.getSharedWorldBounds())) {
					this.visibleObjects.add(sob);
				}
			}
		}
	}

	private final SceneObjectVisitorForVisible sceneObjectVisitorForVisible =
			new SceneObjectVisitorForVisible(this);

	/**
	 * Adds to <code>result</code> all the {@link SceneObject} objects that are
	 * found to be visible by this {@link Camera}.
	 * @param result
	 * @param queueIndex
	 * @return
	 */
	public ArrayList<SceneObject> findVisibleSceneObjects(
			final ArrayList<SceneObject> result, int queueIndex) {
		this.sceneObjectVisitorForVisible.setResultArray(result);
		SceneConductor.getInstance().visitRenderQueue(
				queueIndex, this.sceneObjectVisitorForVisible);
		return result;
	}

	private final ArrayList<SceneObject> visibleSobsRToTex = new ArrayList<SceneObject>();

	/**
	 * Stores SceneObjects and their center in camera space coordinates
	 * so instances of this class can be ordered on axe Z (-Z towards +Z).
	 */
	private static final class OrderedOnZSob implements Comparable<OrderedOnZSob> {
		public SceneObject sob;
		public final Vector3 center = Vector3.getNewZeroVec();
		public boolean used = false;

		/**
		 * Sets the SceneObject that will be sorted and its current center.
		 * It will be needed to transform this center later to make the
		 * method <code>compare</code> useful.
		 * @param sob the SceneObject to render later.
		 * @param center the current center of the SceneObject, not transformed yet.
		 */
		public void setContent(SceneObject sob, Vector3 center) {
			this.sob = sob;
			this.center.setValues(center);
			this.used = true;
		}

		@Override
		public int compareTo(OrderedOnZSob other) {
			final float thisZ = this.center.getZ();
			final float otherZ = other.center.getZ();
			return this.used ? -1 : (thisZ < otherZ ? -1 : (thisZ > otherZ ? 1 : 0));
		}
	};

	private OrderedOnZSob[] orderedSobs = new OrderedOnZSob[0];

	// Used by renderToTexture to avoid many memory allocations.
	private static final Matrix4 modelView_renderToTexture = Matrix4.getNewIdentity();

	public void renderToTexture() {
		preRenderToTexture();

		// Sets the eye position (in word space)
		this.renderingPrms.setEyePosition(getSharedTransform().getSharedWorldTranslation());

		// Sets the view matrix (i.e. inverse of the camera transform).
		getSharedTransform().getInverseWorldMatrix4(this.viewMatrix);
		this.renderingPrms.setViewMatrix(this.viewMatrix);

		this.renderingPrms.setNearAndFarClipPlanes(this.nearClipPlane, this.farClipPlane);

		// Proj x view
		Matrix.multiplyMM(this.viewProjMatrix, 0, this.projMatrix, 0, this.viewMatrix, 0);
		this.renderingPrms.setViewProjMatrix(this.viewProjMatrix);

		this.renderTexture.markAsNotSwapped();

		synchronized (modelView_renderToTexture) {
			getSharedTransform().getInverseWorldMatrix4(modelView_renderToTexture);

			final int nbQueues = SceneConductor.getInstance().countRenderQueues();
			for (int queueIndex = 0; queueIndex < nbQueues; queueIndex++) {
				visibleSobsRToTex.clear();
				findVisibleSceneObjects(visibleSobsRToTex, queueIndex);

				final boolean orderOnZ = (
						SceneConductor.getInstance().getQueueAtIndex(queueIndex) >
						Shader.RenderQueueType.GEOMETRY.index());

				if (!orderOnZ) {
					boolean needsPass = true;
					for (int numPass = 0; needsPass; numPass++) {
						needsPass = false;
						for (int iSob = 0, iSobEnd = visibleSobsRToTex.size(); iSob < iSobEnd; iSob++) {
							final Renderer renderer = visibleSobsRToTex.get(iSob).getSharedRenderer();
							if (numPass == 0) {
								renderer.setToFirstPass();
							}
							needsPass = renderer.renderOnePass(this) || needsPass;
						}
					}
				} else {
					// Create empty slots if needed ('orderedSobs' may need to be enlarged).
					if (visibleSobsRToTex.size() > this.orderedSobs.length) {
						final OrderedOnZSob[] newOrderedSobs = new OrderedOnZSob[visibleSobsRToTex.size()];
						// Copy current content.
						System.arraycopy(this.orderedSobs, 0, newOrderedSobs, 0, this.orderedSobs.length);
						// Creates the slot...
						for (int j = this.orderedSobs.length, jEnd = newOrderedSobs.length; j < jEnd; j++) {
							newOrderedSobs[j] = new OrderedOnZSob();
						}
						// Switch the arrays.
						this.orderedSobs = newOrderedSobs;
					}
					for (int iSob = 0, iSobEnd = visibleSobsRToTex.size(); iSob < iSobEnd; iSob++) {
						final SceneObject sob = visibleSobsRToTex.get(iSob);
						final OrderedOnZSob sobWrapper = this.orderedSobs[iSob];
						sobWrapper.setContent(
								sob,
								sob.getSharedTransform().getSharedWorldTranslation());
						modelView_renderToTexture.transformPoint(
								sobWrapper.center, sobWrapper.center);
					}
					// Sorts from negative to positive numbers, so the Z highest
					// values will appeared last and hence will be the last drawn.
					Arrays.sort(this.orderedSobs);
					// Finally renders.
					boolean needsPass = true;
					for (int numPass = 0; needsPass; numPass++) {
						needsPass = false;
						for (int iSobWrap = 0, iSobWrapEnd = this.orderedSobs.length; iSobWrap < iSobWrapEnd; iSobWrap++) {
							final OrderedOnZSob sobWrapper = this.orderedSobs[iSobWrap];
							if (sobWrapper.used) {
								final Renderer renderer = sobWrapper.sob.getSharedRenderer();
								if (numPass == 0) {
									renderer.setToFirstPass();
								}
								needsPass = renderer.renderOnePass(this) || needsPass;
								// Forget this wrapper;
								sobWrapper.used = false;
							}
						}
					}
				}
			}
		}

		GLES20.glBindFramebuffer(GLES20.GL_FRAMEBUFFER, 0);
		this.renderTexture.swapPrevAndCurrentTexture();
	}

	private void preRenderToTexture() {
		if (this.fbHandle < 1) {
			throw new Utils.WrongFrameBufferHandleException(
					"FBO not initialized? (fbHandle = " + this.fbHandle + ")");
		}

		GLES20.glBindFramebuffer(GLES20.GL_FRAMEBUFFER, this.fbHandle);

		final int renderTextureHandle = this.renderTexture.getTextureHandle();

		GLES20.glFramebufferTexture2D(
				GLES20.GL_FRAMEBUFFER, GLES20.GL_COLOR_ATTACHMENT0,
                GLES20.GL_TEXTURE_2D, renderTextureHandle, 0);

		GLES20.glFramebufferRenderbuffer(
				GLES20.GL_FRAMEBUFFER, GLES20.GL_DEPTH_ATTACHMENT,
				GLES20.GL_RENDERBUFFER, this.depthFbHandle);

		final int fbStatus = GLES20.glCheckFramebufferStatus(GLES20.GL_FRAMEBUFFER);
		if (fbStatus != GLES20.GL_FRAMEBUFFER_COMPLETE) {
			throw new Utils.FrameBufferIncompleteException(
					"Could not bind frame buffer: " + fbStatus +
					", from SceneObject \"" + getSharedSOb().getName() + "\"");
		}

		GLES20.glViewport(0, 0, this.renderTexture.getWidth(), this.renderTexture.getHeight());
		GLES20.glDepthRangef(1.0f, 0.0f);
		// Set the background frame color
		GLES20.glClearColor(
				this.backgroundColor.getR(),
				this.backgroundColor.getG(),
				this.backgroundColor.getB(),
				this.backgroundColor.getA());
		GLES20.glClearDepthf(0.0f);
    	GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT | GLES20.GL_DEPTH_BUFFER_BIT);
	}

	/**
	 * Gets parameters that can be passed to all shaders for a rendering with
	 * this {@link Camera}. Note that info contained by the returned object will
	 * be updated <b>only</b> by the method <code>render</code>.
	 *
	 * @return an object that contains info for rendering.
	 */
	public RenderingValues getSharedRenderingValues() {
		return this.renderingPrms;
	}

	/**
     * Define a projection matrix in terms of a field of view angle, an
     * aspect ratio, and z clip planes.
     * This code has been extracted from Android API level 14 because it is faster
     * than calls to <code>Matrix.frustumM</code>.
     *
     * @param m the float array that holds the perspective matrix
     * @param offset the offset into float array m where the perspective
     * matrix data is written
     * @param fovy field of view in y direction, in degrees
     * @param aspect width to height aspect ratio of the viewport
     * @param zNear a positive number
     * @param zFar a positive number
     */
    public static void perspectiveM(float[] m, int offset,
          float fovy, float aspect, float zNear, float zFar) {
        final float f = 1.0f / (float) Math.tan(fovy * (Math.PI / 360.0));
        final float rangeReciprocal = 1.0f / (zNear - zFar);

        m[offset + 0] = f / aspect;
        m[offset + 1] = 0.0f;
        m[offset + 2] = 0.0f;
        m[offset + 3] = 0.0f;

        m[offset + 4] = 0.0f;
        m[offset + 5] = f;
        m[offset + 6] = 0.0f;
        m[offset + 7] = 0.0f;

        m[offset + 8] = 0.0f;
        m[offset + 9] = 0.0f;
        m[offset + 10] = (zFar + zNear) * rangeReciprocal;
        m[offset + 11] = -1.0f;

        m[offset + 12] = 0.0f;
        m[offset + 13] = 0.0f;
        m[offset + 14] = 2.0f * zFar * zNear * rangeReciprocal;
        m[offset + 15] = 0.0f;
    }

    public void setAlternativePostEffectShader(Shader shader) {
    	this.postEffects.useAlternativeShader(shader);
    }

    public void setToFirstPostEffectPass() {
    	this.postEffects.setToFirstPass();
    }

    public int getCurrentPostEffectPass() {
    	return this.postEffects.getCurrentPass();
    }

    public boolean renderOnePostEffectPass() {
    	return this.postEffects.renderOnePass();
    }

    public boolean postEffectStillNeedsPass() {
    	return this.postEffects.stillNeedsPass();
    }

    // Used by the method 'isVisible' to avoid many memory allocations.
    private static final Vector3 posVec_isVisible = Vector3.getNewZeroVec();
    private static final Vector3 negVec_isVisible = Vector3.getNewZeroVec();
    private static final Vector3 planeNormal_isVisible= Vector3.getNewZeroVec();

    /**
     * Checks that the given AABB, expressed in world coordinates, is visible by
     * this {@link Camera}. An AABB is visible if it intersects at least one
     * of the planes that define the frustum of this {@link Camera}, or if it
     * is inside this frustum.
     *
     * @param bounds the AABB, expressed in world coordinates, to check.
     * @return <code>true</code> if the AABB is visible by this {@link Camera},
     *   <code>false</code> otherwise.
     */
    public boolean isVisible(Bounds bounds) {
    	final int INSIDE = 0;
    	final int INTERSECT = 1;
    	final int OUTSIDE = 2;

    	int state = INSIDE;
    	synchronized (posVec_isVisible) {
    		for (int i = 0; i < this.frustumPlanes.length && state != OUTSIDE; i += 4) {
    			getPosAndNegPointsForPlane(
    					this.frustumPlanes, i, bounds,
    					posVec_isVisible, negVec_isVisible);
    			planeNormal_isVisible.setValues(
    					this.frustumPlanes[i    ],
    					this.frustumPlanes[i + 1],
    					this.frustumPlanes[i + 2]);
    			final float d = this.frustumPlanes[i + 3];

    			final float m = Vector3.dot(planeNormal_isVisible, negVec_isVisible);
    			if (m > -d) {
    				state = OUTSIDE;
    			} else {
    				final float n = Vector3.dot(planeNormal_isVisible, posVec_isVisible);
    				if (n > -d) {
    					state = INTERSECT;
    				}
    			}
    		}
    	}
    	return state == INSIDE || state == INTERSECT;
    }

    private static final int LEFT_PLANE_INDEX = 0;
    private static final int RIGHT_PLANE_INDEX = 4;
    private static final int BOTTOM_PLANE_INDEX = 8;
    private static final int TOP_PLANE_INDEX = 12;
    private static final int NEAR_PLANE_INDEX = 16;
    private static final int FAR_PLANE_INDEX = 20;

    /**
     * Contains sequentially planes that define the current camera frustum:
     * p0a,p0b,p0c,p0d, p1a,p1b,p1c,p1d, ...
     * Each entry contains values of <code>a</code>, <code>b</code>, <code>c</code> and
     * <code>d</code> for equation <code>ax+by+cz+d=0</code>. Equations are
     * expressed in world space.
     *
     * @see LEFT_PLANE_INDEX
     * @see RIGHT_PLANE_INDEX
     * @see BOTTOM_PLANE_INDEX
     * @see TOP_PLANE_INDEX
     * @see NEAR_PLANE_INDEX
     * @see FAR_PLANE_INDEX
     */
    private final float[] frustumPlanes = new float[6 * 4];

    /**
     * Computes the six planes that make up the current camera frustum. The result
     * is stored in the field <code>this.frustumPlanes</code>.
     * The different equations, with form <code>ax+by+cz+d=0</code>, can be found
     * at the following indexes: <code>LEFT_PLANE_INDEX</code>, <code>RIGHT_PLANE_INDEX</code>,
     * <code>BOTTOM_PLANE_INDEX</code>, <code>TOP_PLANE_INDEX</code>,
     * <code>NEAR_PLANE_INDEX</code>, <code>FAR_PLANE_INDEX</code>. Each entry
     * contains values of <code>a</code>, <code>b</code>, <code>c</code> and
     * <code>d</code>.
     */
    private void computeFrustumPlaneEquations() {
    	final Vector3 va = Vector3.getNewZeroVec();
    	final Vector3 vb = Vector3.getNewZeroVec();
    	final Vector3 vn = Vector3.getNewZeroVec();
    	final Vector3 p = Vector3.getNewZeroVec();

    	final float halfHeight = Math.abs(this.nearClipPlane * (float) Math.tan(0.5 * Math.toRadians(this.fovY)));
    	final float halfWidth = halfHeight * this.aspect;
    	final float nearOnFar = this.nearClipPlane / this.farClipPlane;

    	// Left
    	va.setValues(0.0f, halfHeight, 0.0f);
    	vb.setValues(
    			(nearOnFar * (-halfWidth)) - (-halfWidth),
    			(nearOnFar * (-halfHeight)) - (-halfHeight),
    			this.farClipPlane - this.nearClipPlane);
    	Vector3.cross(vn, va, vb);
    	vn.normalizeSelf();
    	p.setValues(-halfWidth, halfHeight, this.nearClipPlane);
    	computePlaneEquation(this.frustumPlanes, LEFT_PLANE_INDEX, vn, p);

    	// Right
    	va.setValues(va.getX(), -va.getY(), va.getZ());
    	// vb remains the same
    	Vector3.cross(vn, va, vb);
    	vn.normalizeSelf();
    	p.setValues(-p.getX(), p.getY(), p.getZ());
    	computePlaneEquation(this.frustumPlanes, RIGHT_PLANE_INDEX, vn, p);

    	// Top
    	va.setValues(halfWidth, 0.0f, 0.0f);
    	// vb remains the same
    	Vector3.cross(vn, va, vb);
    	vn.normalizeSelf();
    	// p remains the same
    	computePlaneEquation(this.frustumPlanes, TOP_PLANE_INDEX, vn, p);

    	// Bottom
    	va.setValues(-va.getX(), va.getY(), va.getZ());
    	// vb remains the same
    	Vector3.cross(vn, va, vb);
    	vn.normalizeSelf();
    	p.setValues(p.getX(), -p.getY(), p.getZ());
    	computePlaneEquation(this.frustumPlanes, BOTTOM_PLANE_INDEX, vn, p);

    	// Near
    	vn.setValues(vb);
    	vn.negateSelf();
    	vn.normalizeSelf();
    	p.setValues(0.0f, 0.0f, this.nearClipPlane);
    	computePlaneEquation(this.frustumPlanes, NEAR_PLANE_INDEX, vn, p);

    	// Far
    	vn.negateSelf();
    	p.setValues(0.0f, 0.0f, this.farClipPlane);
    	computePlaneEquation(this.frustumPlanes, FAR_PLANE_INDEX, vn, p);
    }

    /**
     * Computes the equation as <code>ax+by+cz+d=0</code> of a plane defined by
     * its normal and a point that belongs to this plane.
     *
     * @param result contains sequential planes: p0a,p0b,p0c,p0d, p1a,p1b,p1c,p1d, ...
     *   and also the result located at <code>offset</code>.
     * @param offset is the index that starts equation components of the result plane.
     * @param vn the normal to the plane. It must be <b>normalized</b>.
     * @param point a point that belongs to the plane.
     */
    private void computePlaneEquation(float[] result, int offset, Vector3 vn, Vector3 point) {
    	float d = 0.0f;
    	for (int i = 0; i < 3; i++) {
    		result[offset + i] = vn.getComponent(i);
    		d += result[offset + i] * point.getComponent(i);
    	}
    	result[offset + 3] = -d;
    }

    /**
     * Gets the positive and negative far vertices from bounds (a AABB) against
     * a plane defined by equation <code>ax+by+cz+d=0</code>. The definition used
     * is: the farthest vertex of the AABB in the positive resp. negative sense
     * in the direction of the normal vector of the plane is the <code>p-vertex</code>
     * (the 'positive' far vertex) resp. <code>n-vertex</code> (the 'negative'
     * far vertex).
     *
     * @param planes contains sequential planes: p0a,p0b,p0c,p0d, p1a,p1b,p1c,p1d, ...
     * @param offset is the index that starts equation components of a plane.
     * @param bounds AABB in world coordinates.
     * @param resultPosPoint the 'positive' far vertex that has been found.
     * @param resultNegPoint the 'negative' far vertex that has been found.
     */
    private void getPosAndNegPointsForPlane(
    		float[] planes, int offset, Bounds bounds,
    		Vector3 resultPosPoint, Vector3 resultNegPoint) {
    	final Vector3 vMin = bounds.getSharedMin();
    	final Vector3 vMax = bounds.getSharedMax();
    	final int ltuIndex = (int) (
    			(planes[offset    ] < 0.0f ? 0.0f : 1.0f) * 4.0f +  // a
    			(planes[offset + 1] < 0.0f ? 0.0f : 1.0f) * 2.0f +  // b
    			(planes[offset + 2] < 0.0f ? 0.0f : 1.0f));         // c
    	switch (ltuIndex) {
    	case 7:
    		// + + +
    		resultPosPoint.setValues(vMax);
    		resultNegPoint.setValues(vMin);
    		break;
    	case 6:
    		// + + -
    		resultPosPoint.setValues(vMax.getX(), vMax.getY(), vMin.getZ());
    		resultNegPoint.setValues(vMin.getX(), vMin.getY(), vMax.getZ());
    		break;
    	case 5:
    		// + - +
    		resultPosPoint.setValues(vMax.getX(), vMin.getY(), vMax.getZ());
    		resultNegPoint.setValues(vMin.getX(), vMax.getY(), vMin.getZ());
    		break;
    	case 4:
    		// + - -
    		resultPosPoint.setValues(vMax.getX(), vMin.getY(), vMin.getZ());
    		resultNegPoint.setValues(vMin.getX(), vMax.getY(), vMax.getZ());
    		break;
    	case 3:
    		// - + +
    		resultPosPoint.setValues(vMin.getX(), vMax.getY(), vMax.getZ());
    		resultNegPoint.setValues(vMax.getX(), vMin.getY(), vMin.getZ());
    		break;
    	case 2:
    		// - + -
    		resultPosPoint.setValues(vMin.getX(), vMax.getY(), vMin.getZ());
    		resultNegPoint.setValues(vMax.getX(), vMin.getY(), vMax.getZ());
    		break;
    	case 1:
    		// - - +
    		resultPosPoint.setValues(vMin.getX(), vMin.getY(), vMax.getZ());
    		resultNegPoint.setValues(vMax.getX(), vMax.getY(), vMin.getZ());
    		break;
    	case 0:
    		// - - -
    		resultPosPoint.setValues(vMin);
    		resultNegPoint.setValues(vMax);
    		break;
    	default:
    		resultPosPoint.setValues(Vector3.ZERO);
    		resultNegPoint.setValues(Vector3.ZERO);
    		break;
    	}
    }
}
