package com.laguerre.kan;

import android.opengl.Matrix;

import com.laguerre.kan.RenderingSettings.ShadowSetting;
import com.laguerre.mytest.GaussianBlurImageEffect;

public final class Light extends Behaviour {
	public static enum RenderingMode {
		IMPORTANT,
		NOT_IMPORTANT
	}

	public static enum LightType {
		DIRECTIONAL,
		POINT,
		SPOT
	}

	public static enum ShadowCastingType {
		NONE,
		HARD,
		SOFT
	}

	/**
	 * Contains a default {@link Texture2D} for the depth map. This value will
	 * be initialized by the first call to <code>getDefaultShadowTexture</code>.
	 * By such, it should be possible to build the {@link Texture2D} only when
	 * an OpenGL context is available.
	 */
	private static Texture2D defaultDepthMapTexture = null;

	private static final float[] defaultProjMatrix = makeDefaultProjMatrix();

	private static Texture2D makeDefaultDepthMapTexture() {
		final Texture2D ret = new Texture2D();
		ret.setBitmap(16, 16, Color.BLACK);
		ret.create();
		return ret;
	}

	private static float[] makeDefaultProjMatrix() {
		final float[] ret = new float[16];
		Matrix.orthoM(ret, 0, -0.5f, 0.5f, -0.5f, 0.5f, 0.0f, -1.0f);
		return ret;
	}


	public static Texture2D getDefaultShadowTexture() {
		if (defaultDepthMapTexture == null) {
			defaultDepthMapTexture = makeDefaultDepthMapTexture();
		}
		return defaultDepthMapTexture;
	}

	public static float[] getDefaultProjMatrix() {
		return defaultProjMatrix;
	}

	private LightType lightType = LightType.POINT;
	private RenderingMode renderingMode = RenderingMode.NOT_IMPORTANT;
	private final Color diffuseColor = Color.getNewBlack();
	private final Color specularColor = Color.getNewBlack();
	private float innerCutOffAngleDeg = 10f;
	private float outerCutOffAngleDeg = 15f;
	/**
	 * The reference of this array will be passed and used directly by shaders.
	 * For the sake of performance, the following values are packed here:
	 * <ul>
	 *   <li>cosine inner angle (index 0), computed from <code>innerCutOffAngleDeg</code></li>
	 *   <li>cosine outer angle (index 1), computed from <code>outerCutOffAngleDeg</code></li>
	 * </ul>
	 */
	private final float[] cosineAngleValues = { 0f, 0f };

	/**
	 * The reference of this array will be passed and used directly by shaders.
	 * For the sake of performance, the following values are packed here:
	 * <ul>
	 *   <li>range (index 0)</li>
	 *   <li>constant attenuation (index 1)</li>
	 *   <li>linear attenuation (index 2)</li>
	 *   <li>quadratic attenuation (index 3)</li>
	 * </ul>
	 */
	private final float[] attenuationValues = { -1f, 1f, 0f, 0f };
	private ShadowCastingType shadowType = ShadowCastingType.HARD;
	private SceneObject cameraSob = null;
	private final RenderTexture depthMapRenderTexture = new RenderTexture();
	private Shader shadowReplacementShader = null;

	Light() {
	}

	@Override
	public void onAwake() {
		if (RenderingSettings.getInstance().getShadowSetting() != ShadowSetting.NO_SHADOW) {
			createCameraSob();
		}
	}

	private void createCameraSob() {
		if (this.cameraSob == null) {
			this.cameraSob = SceneConductor.getInstance().addSceneObject(
					"Camera for the shadow map of a light");
			this.cameraSob.getSharedTransform().setParent(getSharedTransform());
			final Camera cam = this.cameraSob.addComponent(Camera.class);
			// Deactivates the camera: only SceneConductor will call its render.
			// method.
			cam.setEnabled(false);
			cam.setAsOrthographic(true);
			// Creates a RenderTexture that will be used as the shadow map.
			this.depthMapRenderTexture.freeResources();
			final int resolution = getResolutionOfShadowMap();
			this.depthMapRenderTexture.init(resolution, resolution);
			this.depthMapRenderTexture.create();
			cam.setRenderTexture(this.depthMapRenderTexture);
			cam.setNearAndFarClipPlanes(
					0.001f,
					RenderingSettings.getInstance().getShadowMaxDistance());
		}
	}

	public float[] getSharedCosineAngleValues() {
		return this.cosineAngleValues;
	}

	public float[] getSharedAttenuationValues() {
		return this.attenuationValues;
	}

	/**
	 * Memory placeholder of the light direction computed by <code>getSharedDirection</code>.
	 * This avoids many memory allocations.
	 */
	private final Vector3 lightDirection = Vector3.getNewOneVec();

	public Vector3 getSharedDirection() {
		// The user is asked to imagine that the ray light goes forward but
		// we use "backward" ray to get an inverse a ray light for shaders.
		getSharedTransform().getSharedWorldRotation().rotateDirection(
				this.lightDirection, Vector3.BACKWARD);
		return this.lightDirection;
	}

	public void setLightType(LightType lightType) {
		this.lightType = lightType;
	}

	public LightType getLightType() {
		return this.lightType;
	}

	public void setRenderingMode(RenderingMode renderingMode) {
		this.renderingMode = renderingMode;
	}

	public RenderingMode getRenderingMode() {
		return this.renderingMode;
	}

	public void setDiffuseColor(float r, float g, float b, float a) {
		this.diffuseColor.setValues(r, g, b, a);
	}

	public void setDiffuseColor(Color c) {
		this.diffuseColor.setValues(c);
	}

	public Color getSharedDiffuseColor() {
		return this.diffuseColor;
	}

	/**
	 * Gets the sum of luminance of the diffuse component and luminance of the
	 * specular component.
	 * @return sum of diffuse and specular luminance values.
	 */
	public float getLuminanceSum() {
		return this.diffuseColor.getLuminance() + this.specularColor.getLuminance();
	}

	public void setSpecularColor(float r, float g, float b, float a) {
		this.specularColor.setValues(r, g, b, a);
	}

	public void setSpecularColor(Color c) {
		this.specularColor.setValues(c);
	}

	public Color getSharedSpecularColor() {
		return this.specularColor;
	}

	public void setRange(float range) {
		this.attenuationValues[0] = range;
	}

	public float getRange() {
		return this.attenuationValues[0];
	}

	public void setConstantAttenuation(float constantAttenuation) {
		this.attenuationValues[1] = constantAttenuation;
	}

	public float getConstantAttenuation() {
		return this.attenuationValues[1];
	}

	public void setLinearAttenuation(float linearAttenuation) {
		this.attenuationValues[2] = linearAttenuation;
	}

	public float getLinearAttenuation() {
		return this.attenuationValues[2];
	}

	public void setQuadraticAttenuation(float quadraticAttenuation) {
		this.attenuationValues[3] = quadraticAttenuation;
	}

	public float getQuadraticAttenuation() {
		return this.attenuationValues[3];
	}

	public void setInnerCutOffAngle(float innerCutOffAngleDeg) {
		this.innerCutOffAngleDeg = innerCutOffAngleDeg;
		this.cosineAngleValues[0] = (float) Math.cos(Math.toRadians(innerCutOffAngleDeg));
	}

	public float getInnerCutOffAngle() {
		return this.innerCutOffAngleDeg;
	}

	public void setOuterCutOffAngle(float outerCutOffAngleDeg) {
		this.outerCutOffAngleDeg = outerCutOffAngleDeg;
		this.cosineAngleValues[1] = (float) Math.cos(Math.toRadians(outerCutOffAngleDeg));
	}

	public float getOuterCutOffAngle() {
		return this.outerCutOffAngleDeg;
	}

	private Component blurPostEffect = null;

	public void setShadow(ShadowCastingType shadowType) {
		if (this.shadowType != shadowType && shadowType != ShadowCastingType.NONE) {
			this.shadowType = shadowType;
			createCameraSob();
			if (shadowType == ShadowCastingType.SOFT) {
				this.blurPostEffect = this.cameraSob.addComponent(
						GaussianBlurImageEffect.class);
			} else {
				if (this.blurPostEffect != null) {
					this.cameraSob.removeComponent(this.blurPostEffect);
					this.blurPostEffect = null;
				}
			}
		}
	}

	public ShadowCastingType getShadow() {
		return this.shadowType;
	}

	void renderToShadowDepthMap() {
		if (this.shadowType != ShadowCastingType.NONE) {
			final Camera cam = this.cameraSob.getSharedCamera();
			cam.setReplacementShader(
					this.shadowReplacementShader, Shader.CommonShaderTags.OPAQUE);
			cam.render();
		}
	}

	public void setShadowReplacementShader(Shader shader) {
		this.shadowReplacementShader = shader;
	}

	public Shader getShadowSharedReplacementShader() {
		return this.shadowReplacementShader;
	}

	public void resetShadowReplacementShader() {
		if (this.cameraSob != null) {
			this.cameraSob.getSharedCamera().resetReplacementShader();
		}
	}

	Texture getDepthMap() {
		return (this.shadowType == ShadowCastingType.NONE ?
				getDefaultShadowTexture() : this.depthMapRenderTexture);
	}

	float[] getSharedProjMatrix() {
		return (this.shadowType == ShadowCastingType.NONE ?
				getDefaultProjMatrix() : this.cameraSob.getSharedCamera().getSharedProjMatrix());
	}

	private int getResolutionOfShadowMap() {
		final int screenWidth = SurfaceInfo.getInstance().getSurfaceWidth();
		final int screenHeight = SurfaceInfo.getInstance().getSurfaceHeight();

		int resolution = (screenWidth > screenHeight ? screenWidth : screenHeight);
		switch (RenderingSettings.getInstance().getShadowResolution()) {
		case LOW:
			resolution *= 0.25;
			break;
		case MEDIUM:
			resolution *= 0.5;
			break;
		case HIGH:
			// nothing to do
			break;
		case VERY_HIGH:
			resolution *= 2;
			break;
		}
		// Limits the size...
		final int maxResolution = 4096;
		if (resolution > maxResolution) {
			resolution = maxResolution;
		}
		return Utils.nextPowerOfTwo(resolution, true);
	}
}
