package com.laguerre.kan;

import java.nio.FloatBuffer;

public final class Material {
	public static interface ShaderVisitor {
		public void visitShader(Shader shader);
	}

	public static class NoShaderSetException extends RuntimeException {
		private static final long serialVersionUID = -3844407651005485335L;

		public NoShaderSetException(String msg) {
			super(msg);
		}
	}

	private String materialName;
	private Shader shader;

	public Material() {
		this.materialName = "";
		this.shader = null;
	}

	public void setMaterialName(String materialName) {
		this.materialName = materialName;
	}

	public String getMaterialName() {
		return this.materialName;
	}

	public void setShader(Shader shader) {
		this.shader = shader;
	}

	public Shader getSharedShader() {
		return this.shader;
	}

	public boolean isShaderSet() {
		return this.shader != null;
	}

	/**
	 * Sets the <code>boolean</code> value associated with <code>variableName</code>
	 * into the provided shader.
	 * @param variableName
	 * @param b
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderValue(String variableName, boolean b) {
		checkShaderIsSet();
		this.shader.setShaderValue(variableName, b);
	}

	/**
	 * Sets the <code>int</code> value associated with <code>variableName</code>
	 * into the provided shader.
	 * @param variableName
	 * @param n
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderValue(String variableName, int n) {
		checkShaderIsSet();
		this.shader.setShaderValue(variableName, n);
	}

	/**
	 * Sets the <code>float</code> value associated with <code>variableName</code>
	 * into the provided shader.
	 * @param variableName
	 * @param f
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderValue(String variableName, float f) {
		checkShaderIsSet();
		this.shader.setShaderValue(variableName, f);
	}

	/**
	 * Sets a reference to an array of <code>boolean</code> values associated
	 * with <code>variableName</code> into the provided shader.
	 * @param variableName
	 * @param refT
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderVector(String variableName, boolean[] refT) {
		checkShaderIsSet();
		this.shader.setShaderVector(variableName, refT);
	}

	/**
	 * Sets a reference to an array of <code>int</code> values associated with
	 * <code>variableName</code> into the provided shader.
	 * @param variableName
	 * @param refT
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderVector(String variableName, int[] refT) {
		checkShaderIsSet();
		this.shader.setShaderVector(variableName, refT);
	}

	/**
	 * Sets the {@link Vector3} values associated with <code>variableName</code>
	 * into the provided shader.
	 * @param variableName
	 * @param v
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderVector(String variableName, Vector3 v) {
		checkShaderIsSet();
		this.shader.setShaderVector(variableName, v);
	}

	/**
	 * Sets the {@link Color} values associated with <code>variableName</code>
	 * into the provided shader.
	 * @param variableName
	 * @param c
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderVector(String variableName, Color c) {
		checkShaderIsSet();
		this.shader.setShaderVector(variableName, c);
	}

	/**
	 * Sets a reference to an array of <code>float</code> values associated with
	 * <code>variableName</code> into the provided shader.
	 * @param variableName
	 * @param refT
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderVector(String variableName, float[] refT) {
		checkShaderIsSet();
		this.shader.setShaderVector(variableName, refT);
	}

	/**
	 * Sets a reference to a matrix 2x2 of <code>float</code> value associated
	 * with <code>variableName</code> into the provided shader.
	 * @param variableName
	 * @param refM
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderMatrix2(String variableName, float[] refM) {
		checkShaderIsSet();
		this.shader.setShaderMatrix2(variableName, refM);
	}

	/**
	 * Sets a reference to a matrix 3x3 of <code>float</code> value associated
	 * with <code>variableName</code> into the provided shader.
	 * @param variableName
	 * @param refM
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderMatrix3(String variableName, float[] refM) {
		checkShaderIsSet();
		this.shader.setShaderMatrix3(variableName, refM);
	}

	/**
	 * Sets a reference to a matrix 4x4 of <code>float</code> value associated
	 * with <code>variableName</code> into the provided shader.
	 * @param variableName
	 * @param refM
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderMatrix4(String variableName, float[] refM) {
		checkShaderIsSet();
		this.shader.setShaderMatrix4(variableName, refM);
	}

	/**
	 * Sets a reference to a {@link Buffer} associated with
	 * <code>variableName</code> into the provided shader.
	 * @param variableName
	 * @param size
	 * @param stride
	 * @param refBuffer
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 * @see Material.releaseShaderBuffer
	 */
	public final void setShaderBuffer(
			String variableName, int size, int stride, FloatBuffer refBuffer) {
		checkShaderIsSet();
		this.shader.setShaderBuffer(variableName, size, stride, refBuffer);
	}

	/**
	 * Sets a reference to a {@link Buffer} associated with
	 * <code>variableName</code> into the provided shader.
	 * @param variableName
	 * @param size
	 * @param stride
	 * @param refBuffer the buffer that owns data to upload to GPU.
	 * @param bufferPosition the position to start reading of data from <code>refBuffer</code>
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 * @see Material.releaseShaderBuffer
	 */
	public final void setShaderBuffer(
			String variableName, int size, int stride, FloatBuffer refBuffer, int bufferPosition) {
		checkShaderIsSet();
		this.shader.setShaderBuffer(variableName, size, stride, refBuffer, bufferPosition);
	}

	public final void createShaderVBO(
			String variableName, int size, int stride,
			FloatBuffer refBuffer, int bufferPosition, Shader.VBOPerfHint perfHint) {
		checkShaderIsSet();
		this.shader.createShaderVBO(variableName, size, stride,
				refBuffer, bufferPosition, perfHint);
	}

	public final void createShaderPackedVBO(
			String packedVBOName, int bufferSize, FloatBuffer refBuffer,
			Shader.VBOPerfHint perfHint) {
		checkShaderIsSet();
		this.shader.createShaderPackedVBO(packedVBOName,
				bufferSize, refBuffer, perfHint);
	}

	public final void identifyDataInPack(String variableOrPackedVBOName,
			String variableName, int bufferSize, int stride, int bufferPosition) {
		checkShaderIsSet();
		this.shader.identifyDataInPack(variableOrPackedVBOName,
				variableName, bufferSize, stride, bufferPosition);
	}

	public final void updateShaderVBO(
			String variableOrPackedVBOName, int size, FloatBuffer refBuffer) {
		checkShaderIsSet();
		this.shader.updateShaderVBO(variableOrPackedVBOName, size, refBuffer);
	}

	public final void updateShaderVBO(String variableOrPackedVBOName) {
		checkShaderIsSet();
		this.shader.updateShaderVBO(variableOrPackedVBOName);
	}

	/**
	 * Releases the buffer associated with <code>variableName</code> into the
	 * provided shader.
	 * @param variableName
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 * @see Material.setShaderFloatBuffer
	 */
	public final void releaseShaderBuffer(String variableName) {
		checkShaderIsSet();
		this.shader.releaseShaderBuffer(variableName);
	}

	/**
	 * Releases all the buffers into the provided shader.
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 * @see Material.releaseShaderBuffer
	 */
	public final void releaseAllShaderBuffers() {
		checkShaderIsSet();
		shader.releaseAllShaderBuffers();
	}

	/**
	 * Sets a reference to a {@link Texture} associated with <code>variableName</code>
	 * into the provided shader.
	 * @param variableName
	 * @param refTexture
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderTexture(String variableName, Texture refTexture) {
		checkShaderIsSet();
		this.shader.setShaderTexture(variableName, refTexture);
	}

	/**
	 * Sets a handle to a {@link Texture} associated with <code>variableName</code>
	 * into the provided shader.
	 * @param variableName
	 * @param refTexture
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void setShaderTexture(String variableName, int textureHandle) {
		checkShaderIsSet();
		this.shader.setShaderTexture(variableName, textureHandle);
	}

	public final boolean shaderVariableExists(String variableName) {
		checkShaderIsSet();
		return this.shader.shaderVariableExists(variableName);
	}

	/**
	 * Removes a variable into the provided shader. Buffers would be released
	 * automatically.
	 * @param variableName
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void removeShaderVariable(String variableName) {
		checkShaderIsSet();
		this.shader.removeShaderVariable(variableName);
	}

	/**
	 * Removes all the variables into the provided shader. Buffers would be
	 * released automatically.
	 * @see Material.setShader
	 * @see Material.isShaderSet
	 */
	public final void removeAllShaderVariables() {
		checkShaderIsSet();
		this.shader.removeAllShaderVariables();
	}

	public final int passCount() {
		int nbPasses = 0;
		if (this.shader != null) {
			nbPasses = this.shader.passCount();
		}
		return nbPasses;
	}

	public void setPass(int passNum) {
		checkShaderIsSet();
		this.shader.setPass(passNum);
	}

	public boolean isPassAvailable(int passNum) {
		boolean ret = false;
		if (this.shader != null) {
			ret = this.shader.isPassAvailable(passNum);
		}
		return ret;
	}

	public void use() {
		checkShaderIsSet();
		this.shader.use();
	}

	private void checkShaderIsSet() {
		if (!isShaderSet()) {
			throw new NoShaderSetException("No shader has been set");
		}
	}
}
