package com.laguerre.kan;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;
import java.util.ArrayList;

public class Mesh extends Component {
	public static final int POSITION_DATA_SIZE = 4;
	public static final int NORMAL_DATA_SIZE = 3;
	public static final int TANGENT_DATA_SIZE = 4;
	public static final int UV_DATA_SIZE = 2;

	protected static final class SubMeshData {
		private final Mesh parentMesh;
		public String name = "";
		public short[] triangles = new short[0];
		public ShortBuffer drawListBuffer;

		SubMeshData(Mesh parentMesh) {
			this.parentMesh = parentMesh;
		}

		public void build() {
	    	final ByteBuffer bbDrawList = ByteBuffer.allocateDirect(triangles.length * Short.SIZE / 8);
	    	bbDrawList.order(ByteOrder.nativeOrder());
	    	this.drawListBuffer = bbDrawList.asShortBuffer();
	    	this.drawListBuffer.put(triangles);
	    	this.drawListBuffer.position(0);
		}

		public void computeBounds(Bounds result) {
			float pMinX = Float.MAX_VALUE;
			float pMinY = Float.MAX_VALUE;
			float pMinZ = Float.MAX_VALUE;
			float pMaxX = Float.MIN_VALUE;
			float pMaxY = Float.MIN_VALUE;
			float pMaxZ = Float.MIN_VALUE;
			for (int i = 0; i < this.triangles.length; i++) {
				final int vertexIndex = this.triangles[i];
				// Min
				if (this.parentMesh.vertices[vertexIndex    ] < pMinX) {
					pMinX = this.parentMesh.vertices[vertexIndex];
				}
				if (this.parentMesh.vertices[vertexIndex + 1] < pMinY) {
					pMinY = this.parentMesh.vertices[vertexIndex + 1];
				}
				if (this.parentMesh.vertices[vertexIndex + 2] < pMinZ) {
					pMinZ = this.parentMesh.vertices[vertexIndex + 2];
				}
				// Max
				if (this.parentMesh.vertices[vertexIndex    ] > pMaxX) {
					pMaxX = this.parentMesh.vertices[vertexIndex];
				}
				if (this.parentMesh.vertices[vertexIndex + 1] > pMaxY) {
					pMaxY = this.parentMesh.vertices[vertexIndex + 1];
				}
				if (this.parentMesh.vertices[vertexIndex + 2] > pMaxZ) {
					pMaxZ = this.parentMesh.vertices[vertexIndex + 2];
				}
			}
			result.setValues(
					// center
					new Vector3(
							(pMaxX + pMinX) / 2.0f,
							(pMaxY + pMinY) / 2.0f,
							(pMaxZ + pMinZ) / 2.0f),
					// size
					new Vector3(
							pMaxX - pMinX,
							pMaxY - pMinY,
							pMaxZ - pMinZ));
		}

		public void recomputeNormals() {
			final int vertexCount = this.parentMesh.vertices.length / POSITION_DATA_SIZE;

			if (this.parentMesh.normals.length / NORMAL_DATA_SIZE != vertexCount) {
				// Not enough place -- or too much -- to store all the coordinates
				// of the normals so we create a new array.
				this.parentMesh.normals = new float[vertexCount * NORMAL_DATA_SIZE];
			}

			final Vector3 p0 = Vector3.getNewZeroVec();
			final Vector3 p1 = Vector3.getNewZeroVec();
			final Vector3 p2 = Vector3.getNewZeroVec();

			final Vector3 v10 = Vector3.getNewZeroVec();
			final Vector3 v12 = Vector3.getNewZeroVec();
			final Vector3 normal = Vector3.getNewZeroVec();

			for (int i = 0, iEnd = this.triangles.length; i < iEnd; i += 3) {
				final int p0Index = this.triangles[i    ];
				final int p1Index = this.triangles[i + 1];
				final int p2Index = this.triangles[i + 2];

				int startVertex = POSITION_DATA_SIZE * p0Index;
				p0.setValues(
						this.parentMesh.vertices[startVertex    ],
						this.parentMesh.vertices[startVertex + 1],
						this.parentMesh.vertices[startVertex + 2]);
				p0.mulSelf(1.0f / this.parentMesh.vertices[startVertex + 3]);
				startVertex = POSITION_DATA_SIZE * p1Index;
				p1.setValues(
						this.parentMesh.vertices[startVertex    ],
						this.parentMesh.vertices[startVertex + 1],
						this.parentMesh.vertices[startVertex + 2]);
				p1.mulSelf(1.0f / this.parentMesh.vertices[startVertex + 3]);
				startVertex = POSITION_DATA_SIZE * p2Index;
				p2.setValues(
						this.parentMesh.vertices[startVertex    ],
						this.parentMesh.vertices[startVertex + 1],
						this.parentMesh.vertices[startVertex + 2]);
				p2.mulSelf(1.0f / this.parentMesh.vertices[startVertex + 3]);

				v10.setValues(p0);
				v10.minusSelf(p1);

				v12.setValues(p2);
				v12.minusSelf(p1);

				Vector3.cross(normal, v12, v10);
				normal.normalizeSelf();

				int startNormal = NORMAL_DATA_SIZE * p0Index;
				this.parentMesh.normals[startNormal    ] = normal.getX();
				this.parentMesh.normals[startNormal + 1] = normal.getY();
				this.parentMesh.normals[startNormal + 2] = normal.getZ();
				startNormal = NORMAL_DATA_SIZE * p1Index;
				this.parentMesh.normals[startNormal    ] = normal.getX();
				this.parentMesh.normals[startNormal + 1] = normal.getY();
				this.parentMesh.normals[startNormal + 2] = normal.getZ();
				startNormal = NORMAL_DATA_SIZE * p2Index;
				this.parentMesh.normals[startNormal    ] = normal.getX();
				this.parentMesh.normals[startNormal + 1] = normal.getY();
				this.parentMesh.normals[startNormal + 2] = normal.getZ();
			}
		}

		public void recomputeTangents() {
			final int vertexCount = this.parentMesh.vertices.length / POSITION_DATA_SIZE;

			if (this.parentMesh.tangents.length / TANGENT_DATA_SIZE != vertexCount) {
				// Not enough place -- or too much -- to store all the coordinates
				// of the tangents so we create a new array.
				this.parentMesh.tangents = new float[vertexCount * TANGENT_DATA_SIZE];
			}

			// Notice: these two creations also set values to 0.
			final float[] tan0 = new float[vertexCount * TANGENT_DATA_SIZE];
			final float[] tan1 = new float[vertexCount * TANGENT_DATA_SIZE];

			final Vector3 p0 = Vector3.getNewZeroVec();
			final Vector3 p1 = Vector3.getNewZeroVec();
			final Vector3 p2 = Vector3.getNewZeroVec();

			final Vector3 tmp0 = Vector3.getNewZeroVec();
			final Vector3 tmp1 = Vector3.getNewZeroVec();

			final Vector3 tmpP0 = Vector3.getNewZeroVec();
			final Vector3 tmpP1 = Vector3.getNewZeroVec();

			final Vector3 sDir = Vector3.getNewZeroVec();
			final Vector3 tDir = Vector3.getNewZeroVec();

			for (int i = 0, iEnd = this.triangles.length; i < iEnd; i += 3) {
				final int p0Index = this.triangles[i    ];
				final int p1Index = this.triangles[i + 1];
				final int p2Index = this.triangles[i + 2];

				int startVertex = POSITION_DATA_SIZE * p0Index;
				p0.setValues(
						this.parentMesh.vertices[startVertex    ],
						this.parentMesh.vertices[startVertex + 1],
						this.parentMesh.vertices[startVertex + 2]);
				p0.mulSelf(1.0f / this.parentMesh.vertices[startVertex + 3]);
				startVertex = POSITION_DATA_SIZE * p1Index;
				p1.setValues(
						this.parentMesh.vertices[startVertex    ],
						this.parentMesh.vertices[startVertex + 1],
						this.parentMesh.vertices[startVertex + 2]);
				p1.mulSelf(1.0f / this.parentMesh.vertices[startVertex + 3]);
				startVertex = POSITION_DATA_SIZE * p2Index;
				p2.setValues(
						this.parentMesh.vertices[startVertex    ],
						this.parentMesh.vertices[startVertex + 1],
						this.parentMesh.vertices[startVertex + 2]);
				p2.mulSelf(1.0f / this.parentMesh.vertices[startVertex + 3]);

				Vector3.minus(tmpP0, p1, p0);
				Vector3.minus(tmpP1, p2, p0);

				final float w0U = this.parentMesh.uvs[UV_DATA_SIZE * p0Index    ];
				final float w0V = this.parentMesh.uvs[UV_DATA_SIZE * p0Index + 1];
				final float w1U = this.parentMesh.uvs[UV_DATA_SIZE * p1Index    ];
				final float w1V = this.parentMesh.uvs[UV_DATA_SIZE * p1Index + 1];
				final float w2U = this.parentMesh.uvs[UV_DATA_SIZE * p2Index    ];
				final float w2V = this.parentMesh.uvs[UV_DATA_SIZE * p2Index + 1];

				final float s0 = w1U - w0U;
				final float s1 = w2U - w0U;
				final float t0 = w1V - w0V;
				final float t1 = w2V - w0V;

			    final float divInR = s0 * t1 - s1 * t0;
			    final float r = (divInR == 0.0f ? 0.0f : 1.0f / divInR);
				// sDir = (t1 * tmp0 - t0 * tmp1) * r
				tmp0.setValues(tmpP0);
				tmp0.mulSelf(t1);
				tmp1.setValues(tmpP1);
				tmp1.mulSelf(t0);
				Vector3.minus(sDir, tmp0, tmp1);
				sDir.mulSelf(r);
				// tDir = (s0 * tmp1 - s1 * tmp0) * r
				tmp0.setValues(tmpP1);
				tmp0.mulSelf(s0);
				tmp1.setValues(tmpP0);
				tmp1.mulSelf(s1);
				Vector3.minus(tDir, tmp0, tmp1);
				tDir.mulSelf(r);

				for (int j = 0; j < 3; j++) {
					tan0[p0Index * TANGENT_DATA_SIZE + j] += sDir.getComponent(j);
					tan0[p1Index * TANGENT_DATA_SIZE + j] += sDir.getComponent(j);
					tan0[p2Index * TANGENT_DATA_SIZE + j] += sDir.getComponent(j);
				}
				for (int j = 0; j < 3; j++) {
					tan1[p0Index * TANGENT_DATA_SIZE + j] += tDir.getComponent(j);
					tan1[p1Index * TANGENT_DATA_SIZE + j] += tDir.getComponent(j);
					tan1[p2Index * TANGENT_DATA_SIZE + j] += tDir.getComponent(j);
				}
			}

			final Vector3 tmpNormal = Vector3.getNewZeroVec();
			final Vector3 tmpTangent = Vector3.getNewZeroVec();
			final Vector3 finalTangent = Vector3.getNewZeroVec();

		    for (int i = 0; i < vertexCount; i++) {
		    	final int startTmpNormal = i * NORMAL_DATA_SIZE;
		    	tmpNormal.setValues(
		    			this.parentMesh.normals[startTmpNormal    ],
		    			this.parentMesh.normals[startTmpNormal + 1],
		    			this.parentMesh.normals[startTmpNormal + 2]);
		    	final int startTmpTangent = i * TANGENT_DATA_SIZE;
		    	tmpTangent.setValues(
		    			tan0[startTmpTangent    ],
		    			tan0[startTmpTangent + 1],
		    			tan0[startTmpTangent + 2]);

		    	// Gram-Schmidt orthogonalize
		        // finalTangent = (t - n * Dot(n, t)).Normalize();
		    	finalTangent.setValues(tmpNormal);
		    	final float dotN_T = Vector3.dot(tmpNormal, tmpTangent);
		    	finalTangent.mulSelf(-dotN_T);
		    	finalTangent.addSelf(tmpTangent);
		    	finalTangent.normalizeSelf();

		    	this.parentMesh.tangents[startTmpTangent    ] = finalTangent.getX();
		    	this.parentMesh.tangents[startTmpTangent + 1] = finalTangent.getY();
		    	this.parentMesh.tangents[startTmpTangent + 2] = finalTangent.getZ();

		        // Calculate handedness
		        // tangent.w = (Dot(Cross(n, t), tan1[i]) < 0.0F) ? -1.0f : 1.0f;
		    	//           = (Dot(tmp0, tmp1)  < 0.0f) ? -1.0f : 1.0f;
		    	Vector3.cross(tmp0, tmpNormal, tmpTangent);
		    	tmp1.setValues(
		    			tan1[startTmpTangent    ],
		    			tan1[startTmpTangent + 1],
		    			tan1[startTmpTangent + 2]);
		    	this.parentMesh.tangents[startTmpTangent + 3] = (
		    			Vector3.dot(tmp0, tmp1) < 0.0f ? -1.0f : 1.0f);
		    }
		}
	}

	private float[] vertices = new float[0];
	private float[] normals = new float[0];
	private float[] tangents = new float[0];
	private float[] uvs = new float[0];
	private FloatBuffer packedData;

	private final ArrayList<SubMeshData> subMeshDataSet = new ArrayList<SubMeshData>();
	private final ArrayList<Bounds> subMeshBounds = new ArrayList<Bounds>();
	private final Bounds bounds = new Bounds(Vector3.ZERO, Vector3.ZERO);
	private boolean beingStatic = true;

	protected final SubMeshData getOrCreateSubMeshData(int subMeshId) {
		for (int i = this.subMeshDataSet.size(), iEnd = subMeshId + 1; i < iEnd; i++) {
			this.subMeshDataSet.add(new SubMeshData(this));
			this.subMeshBounds.add(new Bounds(Vector3.ZERO, Vector3.ZERO));
		}
		return this.subMeshDataSet.get(subMeshId);
	}

	protected static float[] copyToOrCreateArray(float[] srcArray, float[] dstArray) {
		if (dstArray.length != srcArray.length) {
			dstArray = srcArray.clone();
		} else {
			System.arraycopy(srcArray, 0, dstArray, 0, srcArray.length);
		}
		return dstArray;
	}

	protected static short[] copyToOrCreateArray(short[] srcArray, short[] dstArray) {
		if (dstArray.length != srcArray.length) {
			dstArray = srcArray.clone();
		} else {
			System.arraycopy(srcArray, 0, dstArray, 0, srcArray.length);
		}
		return dstArray;
	}

	/**
	 * Gets the lowest id of a sub-mesh with the provided name.
	 * @param name
	 * @return the lowest possible id (it might be <code>0</code>), or
	 *   <code>-1</code> if none was found.
	 */
	public final int getSubMeshIdByName(String name) {
		int ret = -1;
		for (int i = 0, iEnd = this.subMeshDataSet.size(); i < iEnd && ret == -1; i++) {
			if (name == this.subMeshDataSet.get(i).name) {
				ret = i;
			}
		}
		return ret;
	}

	public final void setSubMeshName(String name, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.name = name;
	}

	public final String getSubMeshName(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).name;
	}

	public final void setVertices(float[] vertices) {
		this.vertices = copyToOrCreateArray(vertices, this.vertices);
	}

	public final float[] getSharedVertices() {
		return this.vertices;
	}

	public final float[] getVertices() {
		return this.vertices.clone();
	}

	public final boolean hasNormals() {
		return this.normals.length > 0;
	}

	public final void setNormals(float[] normals) {
		this.normals = copyToOrCreateArray(normals, this.normals);
	}

	public final float[] getSharedNormals() {
		return this.normals;
	}

	public final float[] getNormals(int subMeshId) {
		return this.normals.clone();
	}

	public final boolean hasTangents() {
		return this.tangents.length > 0;
	}

	public final void setTangents(float[] tangents) {
		this.tangents = copyToOrCreateArray(tangents, this.tangents);
	}

	public final float[] getSharedTangents() {
		return this.tangents;
	}

	public final float[] getTangents() {
		return this.tangents.clone();
	}

	public final boolean hasUVs() {
		return this.uvs.length > 0;
	}

	public final void setUVs(float[] uvs) {
		this.uvs = copyToOrCreateArray(uvs, this.uvs);
	}

	public final float[] getSharedUVs() {
		return this.uvs;
	}

	public final float[] getUVs() {
		return this.uvs.clone();
	}

	public final void setTriangles(short[] triangles, int subMeshId) {
		final SubMeshData meshData = getOrCreateSubMeshData(subMeshId);
		meshData.triangles = copyToOrCreateArray(triangles, meshData.triangles);
	}

	public final short[] getSharedTriangles(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).triangles;
	}

	public final short[] getTriangles(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).triangles;
	}

	/**
	 * Recomputes the normals. This method can be used only once the method
	 * <code>setTriangles</code> has been called because it will use triangles.
	 * In addition, this method will take care of creating a new array to store
	 * the coordinates of the normals if the old one does not have the exact size.
	 * @param subMeshId
	 */
	public final void recomputeNormals(int subMeshId) {
		this.subMeshDataSet.get(subMeshId).recomputeNormals();
	}

	/**
	 * Recomputes the tangents. This method can be used only once the methods
	 * <code>setNormals</code>, <code>setUVs</code> and <code>setTriangles</code>
	 * have been called because normals, UV coordinates and triangles are used.
	 * In addition, this method will take care of creating a new array to store
	 * the coordinates of the tangents if the old one does not have the exact size.
	 * @param subMeshId
	 */
	public final void recomputeTangents(int subMeshId) {
		this.subMeshDataSet.get(subMeshId).recomputeTangents();
	}

	public void clear() {
		this.subMeshDataSet.clear();
	}

	public void recomputeNormals() {
		for (int i = 0, iEnd = this.subMeshDataSet.size(); i < iEnd; i++) {
			this.subMeshDataSet.get(i).recomputeNormals();
		}
	}

	public void recomputeTangents() {
		for (int i = 0, iEnd = this.subMeshDataSet.size(); i < iEnd; i++) {
			this.subMeshDataSet.get(i).recomputeTangents();
		}
	}

	public void build() {
		final int vertexCount = vertices.length / POSITION_DATA_SIZE;
		final int dataSize = getVertexPackedDataSize();

    	final float[] vertexData = new float[vertexCount * dataSize];
    	for (int i = 0; i < vertexCount; i++) {
    		int writeCursor = i * dataSize;

    		System.arraycopy(this.vertices, POSITION_DATA_SIZE * i, vertexData, writeCursor, POSITION_DATA_SIZE);
    		writeCursor += POSITION_DATA_SIZE;

    		if (hasNormals()) {
    			System.arraycopy(this.normals, NORMAL_DATA_SIZE * i, vertexData, writeCursor, NORMAL_DATA_SIZE);
	    		writeCursor += NORMAL_DATA_SIZE;
    		}
    		if (hasTangents()) {
    			System.arraycopy(this.tangents, TANGENT_DATA_SIZE * i, vertexData, writeCursor, TANGENT_DATA_SIZE);
	    		writeCursor += TANGENT_DATA_SIZE;
    		}
    		if (hasUVs()) {
    			System.arraycopy(this.uvs, UV_DATA_SIZE * i, vertexData, writeCursor, UV_DATA_SIZE);
	    		writeCursor += UV_DATA_SIZE;
    		}
    	}

    	final ByteBuffer bb = ByteBuffer.allocateDirect(vertexCount * dataSize * Float.SIZE / 8);
    	bb.order(ByteOrder.nativeOrder());
    	this.packedData = bb.asFloatBuffer();
    	this.packedData.put(vertexData);
    	this.packedData.position(0);

    	// Resets the bounds to recompute them.
    	this.bounds.setValues(Vector3.ZERO, Vector3.ZERO);

    	for (int i = 0, iEnd = this.subMeshDataSet.size(); i < iEnd; i++) {
    		final SubMeshData subMeshData = this.subMeshDataSet.get(i);
    		subMeshData.build();

    		final Bounds subMeshBounds = this.subMeshBounds.get(i);
    		subMeshData.computeBounds(subMeshBounds);
    		this.bounds.encapsulate(subMeshBounds);
    	}
	}

	public final int getNbOfSubMeshes() {
		return this.subMeshDataSet.size();
	}

	public int getVertexPackedDataSize() {
		int dataSize = POSITION_DATA_SIZE;
		if (hasNormals()) {
			dataSize += NORMAL_DATA_SIZE;
		}
		if (hasTangents()) {
			dataSize += TANGENT_DATA_SIZE;
		}
		if (hasUVs()) {
			dataSize += UV_DATA_SIZE;
		}
		// Adds padding if needed to reach a multiple of 4 bytes.
		dataSize += (dataSize % 4);
		return dataSize;
	}

	public final int getVertexCount() {
		return this.vertices.length / POSITION_DATA_SIZE;
	}

	public final int getVertexCoordinatesCount() {
		return this.vertices.length;
	}

	public final FloatBuffer getSharedVertexPackedData() {
		return this.packedData;
	}

	public final int getDrawListCount(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).triangles.length;
	}

	public final ShortBuffer getSharedDrawListBuffer(int subMeshId) {
		return this.subMeshDataSet.get(subMeshId).drawListBuffer;
	}

	public final Bounds getSharedBounds() {
		return this.bounds;
	}

	public final Bounds getBounds() {
		return new Bounds(this.bounds);
	}

	public final void markAsDynamic() {
		this.beingStatic = false;
	}

	public final void markAsStatic() {
		this.beingStatic = true;
	}

	public final boolean isStatic() {
		return this.beingStatic;
	}
}
