package com.laguerre.kan;

import java.nio.FloatBuffer;
import java.util.ArrayList;

import android.util.SparseArray;

public class MeshRenderer extends Renderer {
	private final SparseArray<Material> materialPerSubMesh = new SparseArray<Material>();
	protected int passNum = 0;
	protected boolean needsPass = true;
	private final ArrayList<TriangleEBO> triangleEBOs = new ArrayList<TriangleEBO>();

	public final void setMaterial(Material mat) {
		setMaterial(mat, 0);
	}

	@Override
	public final void setMaterial(Material mat, int subMeshId) {
		this.materialPerSubMesh.put(subMeshId, mat);
	}

	@Override
	public final Material getSharedMaterial(int subMeshId) {
		return this.materialPerSubMesh.get(subMeshId);
	}

	@Override
	public int findSubMeshIdOfMaterial(String materialName) {
		int ret = -1;
		for (int i = 0, iEnd = this.materialPerSubMesh.size(); i < iEnd && ret == -1; i++) {
			final Material mat = this.materialPerSubMesh.valueAt(i);
			if (mat != null && mat.getMaterialName().equals(materialName)) {
				ret = this.materialPerSubMesh.keyAt(i);
			}
		}
		return ret;
	}

	@Override
	public Material findMaterial(String materialName) {
		Material ret = null;
		final int subMeshId = findSubMeshIdOfMaterial(materialName);
		if (subMeshId != -1) {
			ret = this.materialPerSubMesh.get(subMeshId);
		}
		return ret;
	}

	@Override
	public void setToFirstPass() {
		this.passNum = 0;
		this.needsPass = true;
	}

	@Override
	public int getCurrentPass() {
		return this.passNum;
	}

	/**
	 * Shader values that are common to every shader of this {@link SceneObject}.
	 */
	private final Shader.ModelRelativeData shaderValues = new Shader.ModelRelativeData();

	// Used only by renderOnePass() to avoid many memory allocations.
	private static final float[] renderBufferMat4 = new float[16];
	private static final float[] renderBufferInvMat4 = new float[16];

	@Override
	public boolean renderOnePass(Camera camera) {
		if (!this.needsPass) {
			return false;
		}

		// will be updated once a rendering occurred.
		this.needsPass = false;

		final SceneObject parentSOb = getSharedSOb();
		final Mesh mesh = parentSOb.getAnyComponent(Mesh.class);
		if (mesh != null) {
			final Transform t = parentSOb.getSharedTransform();

			// world-to-local and local-to-world matrices
			shaderValues.setWorldToLocalMatrix(
					t.getSharedWorldToLocalMatrix().getSharedComponents(),
					t.getSharedLocalToWorldMatrix().getSharedComponents());

			// Model matrix and the inverse
			shaderValues.setModelMatrix(
					t.getSharedWorldMatrix4().getSharedComponents(),
					t.getSharedInverseWorldMatrix4().getSharedComponents());

			// Other shader values that are camera dependent only.
			final Camera.RenderingValues cameraVals = camera.getSharedRenderingValues();

			synchronized (renderBufferMat4) {
				// View-model matrix and the inverse
				Transform.combine(renderBufferMat4, cameraVals.getSharedViewMatrix(), shaderValues.getSharedModelMatrix());
				Transform.combine(renderBufferInvMat4, shaderValues.getSharedInvModelMatrix(), cameraVals.getSharedInvViewMatrix());
				shaderValues.setModelViewMatrix(renderBufferMat4, renderBufferInvMat4);

				// Proj-view-model matrix and the inverse
				Transform.combine(renderBufferMat4, cameraVals.getSharedProjMatrix(), shaderValues.getSharedModelViewMatrix());
				Transform.combine(renderBufferInvMat4, shaderValues.getSharedInvModelViewMatrix(), cameraVals.getSharedInvProjMatrix());
				shaderValues.setModelViewProjModelMatrix(renderBufferMat4, renderBufferInvMat4);
			}

			final int byteStride = mesh.getVertexPackedDataSize() * Float.SIZE / 8;
			final FloatBuffer packedDataBuffer = mesh.getSharedVertexPackedData();

			final String replacementTag = camera.getSharedReplacementTag();

			final int nbOfSubMeshes = mesh.getNbOfSubMeshes();
			for (int i = 0; i < nbOfSubMeshes; i++) {
				final Material mat = this.materialPerSubMesh.get(i);
				if (mat != null) {
					// Initialize the texture counter for all the Materials of this sub-mesh.
					Texture.resetTextureCounter();

					if (mat.isPassAvailable(this.passNum)) {
						final Shader originalShader = mat.getSharedShader();
						final Shader replacementShader = camera.getSharedReplacementShader();

						if (replacementShader == null ||
								(replacementTag == null || replacementTag.length() == 0 ||
								originalShader.getTagValue(
										Shader.CommonShaderTagTypes.RENDER_TYPE).equals(replacementTag))) {

							if (replacementShader != null) {
								mat.setShader(replacementShader);
							}

							mat.setShaderVector("viewSize", cameraVals.getSharedViewSize());
							//
							mat.setShaderVector("nearAndFarClipPlane", cameraVals.getSharedNearAndFarClipPlanes());
							//
							mat.setShaderVector("eyePosition", cameraVals.getSharedEyePosition());
							//
							//getSharedWorldToLocalMatrix
							mat.setShaderMatrix4("WorldToLocalMatrix", shaderValues.getSharedWorldToLocalMatrix());
							mat.setShaderMatrix4("LocalToWorldMatrix", shaderValues.getSharedLocalToWorldMatrixMatrix());
							//
							mat.setShaderMatrix4("ViewMatrix", cameraVals.getSharedViewMatrix());
							mat.setShaderMatrix4("invViewMatrix", cameraVals.getSharedInvViewMatrix());
							//
							mat.setShaderMatrix4("ProjMatrix", cameraVals.getSharedProjMatrix());
							mat.setShaderMatrix4("invProjMatrix", cameraVals.getSharedInvProjMatrix());
							//
							mat.setShaderMatrix4("VPMatrix", cameraVals.getSharedViewProjMatrix());
							mat.setShaderMatrix4("invVPMatrix", cameraVals.getSharedInvViewProjMatrix());
							//
							mat.setShaderMatrix4("ModelMatrix", shaderValues.getSharedModelMatrix());
							mat.setShaderMatrix4("invModelMatrix", shaderValues.getSharedInvModelMatrix());
							//
							mat.setShaderMatrix4("MVMatrix", shaderValues.getSharedModelViewMatrix());
							mat.setShaderMatrix4("invMVMatrix", shaderValues.getSharedInvModelViewMatrix());
							//
							mat.setShaderMatrix4("MVPMatrix", shaderValues.getSharedModelViewProjMatrix());
							mat.setShaderMatrix4("invMVPMatrix", shaderValues.getSharedInvModelViewProjMatri());
							//
							final String packedVBOName = "packedVBO";
							if (!mat.shaderVariableExists(packedVBOName)) {
								mat.createShaderPackedVBO(
										packedVBOName,
										packedDataBuffer.capacity() * Float.SIZE / 8,
										packedDataBuffer,
										mesh.isStatic() ? Shader.VBOPerfHint.STATIC_DRAW : Shader.VBOPerfHint.DYNAMIC_DRAW);
								mat.identifyDataInPack(packedVBOName, "position", Mesh.POSITION_DATA_SIZE, byteStride, 0);
								int readCursor = Mesh.POSITION_DATA_SIZE;
								if (mesh.hasNormals()) {
									mat.identifyDataInPack(packedVBOName, "normal", Mesh.NORMAL_DATA_SIZE, byteStride, readCursor);
									readCursor += Mesh.NORMAL_DATA_SIZE;
								}
								if (mesh.hasTangents()) {
									mat.identifyDataInPack(packedVBOName, "tangent", Mesh.TANGENT_DATA_SIZE, byteStride, readCursor);
									readCursor += Mesh.TANGENT_DATA_SIZE;
								}
								if (mesh.hasUVs()) {
									mat.identifyDataInPack(packedVBOName, "uv", Mesh.UV_DATA_SIZE, byteStride, readCursor);
									readCursor += Mesh.UV_DATA_SIZE;
								}
							} else {
								mat.updateShaderVBO(
										packedVBOName,
										packedDataBuffer.capacity() * Float.SIZE / 8,
										packedDataBuffer);
							}
							//
							mat.setShaderValue(
									"hasRenderedTex",
									camera.getSharedRenderTexture().hasSwapped());
							mat.setShaderTexture(
									"prevRenderedTex",
									camera.getSharedRenderTexture().getPrevTextureHandle());

							mat.setShaderVector("ambientLight",
									RenderingSettings.getInstance().getSharedAmbientLight());

							Lights.getInstance().passLightsToShader(camera, mat.getSharedShader());

							mat.setShaderValue("isCastingShadows", this.castingShadows);
							mat.setShaderValue("isReceivingShadows", this.receivingShadows);

							mat.setPass(this.passNum);

							mat.use();

							final TriangleEBO triangleVBO = getTriangleEBO(i);
							if (triangleVBO.isNew()) {
								triangleVBO.createData(
										mesh.getSharedDrawListBuffer(i),
										(mesh.isStatic() ? TriangleEBO.EBOPerfHint.STATIC_DRAW :
											TriangleEBO.EBOPerfHint.DYNAMIC_DRAW));
							} else if (!mesh.isStatic()) {
								triangleVBO.updateData(mesh.getSharedDrawListBuffer(i));
							}
							triangleVBO.drawElements();

							mat.releaseAllShaderBuffers();

							if (replacementShader != null) {
								mat.setShader(originalShader);
							}
						}

						// Still passes required?
						final int nbPasses = mat.passCount();
						// 1st pass = 0!
						if (this.passNum + 1 < nbPasses) {
							this.needsPass = true;
						}
					}
				}
			}
			// Removes EBO that are not in use anymore (if any).
			compactNbTriangleEBOs(nbOfSubMeshes);
		}

		// Prepare for the next run.
		this.passNum++;

		return this.needsPass;
	}

	private TriangleEBO getTriangleEBO(int subMeshId) {
		// Creates as much triangles as needed to get a placeholder for the
		// given sub-mesh id.
		for (int i = this.triangleEBOs.size() - 1; i < subMeshId; i++) {
			this.triangleEBOs.add(new TriangleEBO());
		}
		return this.triangleEBOs.get(subMeshId);
	}

	private void compactNbTriangleEBOs(int nbSubMeshes) {
		for (int i = nbSubMeshes, iEnd = this.triangleEBOs.size(); i < iEnd; i++) {
			final TriangleEBO triEBO = this.triangleEBOs.remove(this.triangleEBOs.size() - 1);
			// Free resources
			triEBO.freeResources();
		}
	}

	@Override
	public boolean stillNeedsPass() {
		return this.needsPass;
	}

	@Override
	void updateWorldBounds() {
		final Mesh mesh = getSharedSOb().getAnyComponent(Mesh.class);
		if (mesh != null) {
			updateWorldBounds(mesh.getSharedBounds());
		}
	}
}
