package com.laguerre.kan;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.HashMap;

import com.laguerre.kan.MeshImporter;
import com.laguerre.kan.MeshImporter.Settings;

public class ObjImporter implements MeshImporter.IMeshImporterImpl {
	public static class UnexpectedNumberOfValuesException extends RuntimeException {
		private static final long serialVersionUID = 6867346167225379981L;

		public UnexpectedNumberOfValuesException(String msg) {
			super(msg);
		}
	}

	public static class TooHightIndexException extends RuntimeException {
		private static final long serialVersionUID = -7506392685572998497L;

		public TooHightIndexException(String msg) {
			super(msg);
		}
	}

	private Settings settings;

	private final HashMap<String, Integer> vertexDataIndexes = new HashMap<String, Integer>();

	private void clearVertexDataIndexes() {
		this.vertexDataIndexes.clear();
	}

	private int getVertexDataIndex(int vertexIndex, int normalIndex, int uvIndex) {
		final String key = vertexIndex + "-" + normalIndex + "-" + uvIndex;
		int ret = -1;
		final Integer index = this.vertexDataIndexes.get(key);
		if (index != null) {
			ret = index.intValue();
		}
		return ret;
	}

	private void addVertexDataIndex(int newVertexIndex, int vertexIndex, int normalIndex, int uvIndex) {
		final String key = vertexIndex + "-" + normalIndex + "-" + uvIndex;
		this.vertexDataIndexes.put(key, newVertexIndex);
	}


	@Override
	public void setSettings(Settings settings) {
		this.settings = settings;
	}

	@Override
	public void importToTarget(InputStream inputStream, SceneObject sobTarget) {
		Mesh mesh = sobTarget.getAnyComponent(Mesh.class);
		if (mesh == null) {
			 mesh = sobTarget.addComponent(Mesh.class);
		}
		loadContent(inputStream, mesh);
	}

	protected void loadContent(InputStream inputStream, Mesh mesh) {
		final InputStreamReader inputStreamReader = new InputStreamReader(
				inputStream);
		final BufferedReader bufferedReader = new BufferedReader(inputStreamReader, 8192);
		try {
			final Renderer renderer = mesh.getSharedRenderer();

			// Removes any previous data that represents this mesh.
			mesh.clear();

			// Will contain read data.
			final ArrayList<Float> readVertexData = new ArrayList<Float>();
			final ArrayList<Float> readNormalData = new ArrayList<Float>();
			final ArrayList<Float> readUVData = new ArrayList<Float>();

			// Will contain data built dynamically
			String subMeshName = "";
			int subMeshId = 0;
			// Below, each index contains values for the same vertex (coordinates,
			// normals and uv).
			final ArrayList<Float> finalVertexData = new ArrayList<Float>();
			final ArrayList<Float> finalNormalData = new ArrayList<Float>();
			final ArrayList<Float> finalUVData = new ArrayList<Float>();
			final ArrayList<Integer> finalFaceData = new ArrayList<Integer>();

			clearVertexDataIndexes();

			String currentLine = null;
			while ((currentLine = bufferedReader.readLine()) != null) {
				currentLine = currentLine.trim();
				final String[] lineContent = currentLine.split("\\s+");

				if (lineContent.length == 0 || lineContent[0].startsWith("#")) {
					// Nothing to do...
					continue;
				} else if (lineContent[0].equals("v")) {
					if (lineContent.length < 4) {
						throw new UnexpectedNumberOfValuesException("A vertex did not provide at least 3 values");
					}
					for (int i = 1; i < 4; i++) {
						readVertexData.add(this.settings.globalScale * Float.parseFloat(lineContent[i]));
					}
					if (lineContent.length >= 5) {
						// "w" is provided.
						readVertexData.add(Float.parseFloat(lineContent[4]));
					} else {
						// Sets "w" as 1.0
						readVertexData.add(1.0f);
					}
					// Assumes that vertices was the first data to be loaded so
					// memory is reserved for further loading of normals and UV.
					// It will be a waste if no normals are provided but much
					// faster in the opposite case...
					final int nbVertices = readVertexData.size() / 4;
					readNormalData.ensureCapacity(nbVertices * 3);
					readUVData.ensureCapacity(nbVertices * 2);
				} else if (lineContent[0].equals("vn")) {
					if (lineContent.length != 4) {
						throw new UnexpectedNumberOfValuesException("A normal did not provide 3 values");
					}
					for (int i = 1; i < 4; i++) {
						readNormalData.add(Float.parseFloat(lineContent[i]));
					}
				} else if (lineContent[0].equals("vt")) {
					if (lineContent.length != 3 && lineContent.length != 4) {
						throw new UnexpectedNumberOfValuesException("A UV did not provide 2 values");
					}
					if (lineContent.length == 4) {
						final float tCoord = Float.parseFloat(lineContent[3]);
						if (!Utils.fuzzyEquals(0.0f, tCoord)) {
							throw new UnexpectedNumberOfValuesException("UV only supports 2 values, or the third must equal to 0");
						}
					}
					for (int i = 1; i < 3; i++) {
						readUVData.add(Float.parseFloat(lineContent[i]));
					}
					if (this.settings.swappedUV) {
						final int readUVDataSize = readUVData.size();
						final float oldUComponent = readUVData.get(readUVDataSize - 2);
						readUVData.set(readUVDataSize - 2, readUVData.get(readUVDataSize - 1));
						readUVData.set(readUVDataSize - 1, oldUComponent);
					}
				} else if (lineContent[0].equals("f")) {
					if (lineContent.length < 4) {
						throw new UnexpectedNumberOfValuesException("A face did not provide at least 3 vertices");
					}

					for (int i = 2; i + 1 < lineContent.length; i++) {
						// Adds p (1), it is the beginning of a triangle fan.
						addVertexAndFaceData(
								lineContent[1],
								readVertexData,
								readNormalData,
								readUVData,
								finalVertexData,
								finalNormalData,
								finalUVData,
								finalFaceData);

						// Adds p (i)
						addVertexAndFaceData(
								lineContent[i],
								readVertexData,
								readNormalData,
								readUVData,
								finalVertexData,
								finalNormalData,
								finalUVData,
								finalFaceData);

						// Adds p (i + 1)
						addVertexAndFaceData(
								lineContent[i + 1],
								readVertexData,
								readNormalData,
								readUVData,
								finalVertexData,
								finalNormalData,
								finalUVData,
								finalFaceData);
					}
				} else if (lineContent[0].equals("g") ||
						lineContent[0].equals("usemtl")) {
					// Any mesh data loaded previously? If true, builds the mesh.
					if (!finalFaceData.isEmpty()) {
						setSubMeshData(mesh, subMeshId, subMeshName, finalFaceData);
						subMeshId++;
						// Ensures that we aren't keeping garbage.
						finalFaceData.clear();
						clearVertexDataIndexes();
					}

					if (lineContent[0].equals("g")) {
						// The name of the new mesh we're beginning to load.
						if (lineContent.length > 1) {
							// Number 2 for "g ".
							subMeshName = currentLine.substring(2).trim();
						}
					} else {
						// Material name is declared.
						if (renderer != null && lineContent.length > 1) {
							Material matSubMesh = renderer.getSharedMaterial(subMeshId);
							if (matSubMesh == null) {
								matSubMesh = new Material();
								renderer.setMaterial(matSubMesh, subMeshId);
							}
							// Number 7 for "usemtl ".
							final String materialName = currentLine.substring(7).trim();
							matSubMesh.setMaterialName(materialName);
						}
					}
				}
			}

			// Treats the last defined sub-mesh.
			if (!finalFaceData.isEmpty()) {
				setSubMeshData(mesh, subMeshId, subMeshName, finalFaceData);
			}

			setGlobalMeshData(mesh, finalVertexData, finalNormalData, finalUVData);
			// And builds the mesh...
			mesh.build();

		} catch (IOException e) {
		}
	}

	protected void addVertexAndFaceData(
			final String vertexDataStr,
			final ArrayList<Float> readVertexData,
			final ArrayList<Float> readNormalData,
			final ArrayList<Float> readUVData,
			final ArrayList<Float> finalVertexData,
			final ArrayList<Float> finalNormalData,
			final ArrayList<Float> finalUVData,
			final ArrayList<Integer> finalFaceData) {
		final String[] faceContent = vertexDataStr.split("/");
		boolean onlyVertexIndex = false;
		boolean normalProvided = false;
		boolean uvProvided = false;
		int vertexIndex;
		int normalIndex = -1 ;
		int uvIndex = -1;
		if (faceContent.length == 1) {
			onlyVertexIndex = true;
			// Only vertex indexes are given.
			vertexIndex = Integer.parseInt(vertexDataStr);
		} else {
			if (faceContent.length == 2) {
				// Format: v/vt.
				vertexIndex = Integer.parseInt(faceContent[0]);
				// Always a normal.
				uvIndex = Integer.parseInt(faceContent[1]);
				uvProvided = true;
			} else {
				// Format: v/vt/vn.
				// An uv can be ignored with a slash, e.g v//vn.
				vertexIndex = Integer.parseInt(faceContent[0]);
				// Always a normal.
				normalIndex = Integer.parseInt(faceContent[2]);
				normalProvided = true;
				// UV also?
				if (faceContent[1].length() > 0) {
					uvIndex = Integer.parseInt(faceContent[1]);
					uvProvided = true;
				}
			}
		}
		if (vertexIndex < 0) {
			// negative means that it is a reference to the just-added vertices.
			vertexIndex = readVertexData.size() + vertexIndex;
		} else {
			// First index is 0 for us, not 1.
			vertexIndex--;
		}

		if (onlyVertexIndex) {
			// No index provided? But maybe some have been loaded
			// and thus index is also the vertex index...
			final int tmpNormalIndex = vertexIndex * Mesh.NORMAL_DATA_SIZE;
			if (tmpNormalIndex < readNormalData.size() - 1 - Mesh.NORMAL_DATA_SIZE) {
				// ok, a normal is available
				normalIndex = vertexIndex;
			}
			// And for UV?
			final int tmpUVIndex = vertexIndex * Mesh.UV_DATA_SIZE;
			if (tmpUVIndex < readUVData.size() - 1 - Mesh.UV_DATA_SIZE) {
				// UV is available
				uvIndex = vertexIndex;
			}
		} else {
			// Copies normal coordinates, if any.
			if (normalProvided) {
				// First index is 0 for us, not 1.
				normalIndex--;
			}
			// Copies uv coordinates, if any.
			if (uvProvided) {
				// First index is 0 for us, not 1.
				uvIndex--;
			}
		}

		// Recycle known-vertex data or make new ones.
		int finalVertexIndex = getVertexDataIndex(vertexIndex, normalIndex, uvIndex);
		if (finalVertexIndex == -1) {
			// A new tuple so a new vertex has to be created.
			//
			// Hence a copy of vertex coordinates.
			for (int i = vertexIndex * Mesh.POSITION_DATA_SIZE, iEnd = i + Mesh.POSITION_DATA_SIZE; i < iEnd; i++) {
				finalVertexData.add(readVertexData.get(i));
			}
			// Normal coordinates...
			for (int i = normalIndex * Mesh.NORMAL_DATA_SIZE, iEnd = i + Mesh.NORMAL_DATA_SIZE; i < iEnd; i++) {
				finalNormalData.add(readNormalData.get(i));
			}
			// Finally UV coordinates.
			for (int i = uvIndex * Mesh.UV_DATA_SIZE, iEnd = i + Mesh.UV_DATA_SIZE; i < iEnd; i++) {
				finalUVData.add(readUVData.get(i));
			}
			// Stores this tuple.
			finalVertexIndex = finalVertexData.size() / Mesh.POSITION_DATA_SIZE - 1;
			addVertexDataIndex(finalVertexIndex, vertexIndex, normalIndex, uvIndex);
		}

		finalFaceData.add(finalVertexIndex);
	}

	protected void setSubMeshData(
			final Mesh mesh,
			final int subMeshId,
			final String subMeshName,
			final ArrayList<Integer> faceData) {
		short[] shortData = new short[faceData.size()];
		for (int i = 0, iEnd = faceData.size(); i < iEnd; i++) {
			//
			final int vertexIndex = faceData.get(i);
			if (vertexIndex > Short.MAX_VALUE) {
				throw new TooHightIndexException("Got index \"" + vertexIndex +
						"\", which is too big (> " + Short.MAX_VALUE + ")");
			}
			shortData[i] = (short) vertexIndex;
		}
		mesh.setTriangles(shortData, subMeshId);
	}

	protected void setGlobalMeshData(
			final Mesh mesh,
			final ArrayList<Float> vertexData,
			final ArrayList<Float> normalData,
			final ArrayList<Float> uvData) {
		float[] floatData = new float[vertexData.size()];
		for (int i = 0, iEnd = vertexData.size(); i < iEnd; i++) {
			floatData[i] = vertexData.get(i);
		}
		mesh.setVertices(floatData);

		switch (this.settings.normalImportMode) {
		case IMPORT_FROM_FILE:
			if (!normalData.isEmpty()) {
				floatData = new float[normalData.size()];
				for (int i = 0, iEnd = normalData.size(); i < iEnd; i++) {
					floatData[i] = normalData.get(i);
				}
				mesh.setNormals(floatData);
			}
			break;
		case CALCULATE:
			mesh.recomputeNormals();
			break;
		case NONE:
		default:
			break;
		}

		if (!uvData.isEmpty()) {
			floatData = new float[uvData.size()];
			for (int i = 0, iEnd = uvData.size(); i < iEnd; i++) {
				floatData[i] = uvData.get(i);
			}
			mesh.setUVs(floatData);
		}

		switch (this.settings.tangentImportMode) {
		case IMPORT_FROM_FILE:
			// Not supported by this file format.
			break;
		case CALCULATE:
			mesh.recomputeTangents();
			break;
		case NONE:
		default:
			break;
		}
	}
}
