package com.laguerre.kan;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;

import com.laguerre.kan.ParticleSystem.Particle;

public final class ParticleSystemRenderer extends Renderer {
	private Material material;
	private int passNum = 0;
	private TriangleEBO triangleEBO;

	@Override
	public void setMaterial(Material mat, int subMeshId) {
		this.material = mat;
	}

	@Override
	public Material getSharedMaterial(int subMeshId) {
		return this.material;
	}

	@Override
	public int findSubMeshIdOfMaterial(String materialName) {
		final Material mat = findMaterial(materialName);
		return (mat != null ? 0 : -1);
	}

	@Override
	public Material findMaterial(String materialName) {
		Material ret = null;
		if (material != null && materialName.equals(material.getMaterialName())) {
			ret = this.material;
		}
		return ret;
	}

	@Override
	public void setToFirstPass() {
		this.passNum = 0;
	}

	@Override
	public int getCurrentPass() {
		return this.passNum;
	}

	private static final int POSITION_DATA_SIZE = 4;
	private static final int COLOR_DATA_SIZE = 4;
	private static final int UV_DATA_SIZE = 2;
	/** Adds bytes to get finally 4-bytes alignment (and better performances). */
	private static final int PADDING_DATA_SIZE = 2;
	/**
	 * Vertex packed data with the following contain, in this order:
	 * 1/ vertex position (x, y, z, w) -- float
	 * 2/ vertex color (r, g, b, a) -- float
	 * 3/ vertex texture coordinates (u, v) -- float.
	 *
	 * In addition, 4 vertices are created per particle:
	 * p0,p1,p2,p3,p0,p1,p2,p3, etc., which leads to:
	 * xyzwrgbauv,xyzwrgbauv,xyzwrgbauv,xyzwrgbauv|...
	 */
	private FloatBuffer packedDataBuffer;

	private ShortBuffer drawListBuffer;

	// Used only by renderOnePass() to avoid many memory allocations.
	private static final float[] renderBufferMat4 = new float[16];
	private static final float[] renderBufferInvMat4 = new float[16];

	/**
	 * Shader values that are common to every shader of this {@link SceneObject}.
	 */
	private final Shader.ModelRelativeData shaderValues = new Shader.ModelRelativeData();

	@Override
	public boolean renderOnePass(Camera camera) {
		final Shader originalShader = this.material.getSharedShader();

		final String replacementTag = camera.getSharedReplacementTag();
		final Shader replacementShader = camera.getSharedReplacementShader();

		if (replacementShader == null ||
				(replacementTag == null || replacementTag.length() == 0 ||
				originalShader.getTagValue(
						Shader.CommonShaderTagTypes.RENDER_TYPE).equals(replacementTag))) {

			final SceneObject parentSOb = getSharedSOb();
			final ParticleSystem particleSystem = parentSOb.getAnyComponent(ParticleSystem.class);
			if (particleSystem != null && particleSystem.countAliveParticles() > 0) {
				final Transform t = parentSOb.getSharedTransform();

				synchronized (renderBufferMat4) {
					// Model matrix and the inverse
					t.getWorldMatrix4(renderBufferMat4);
					t.getInverseWorldMatrix4(renderBufferInvMat4);
					shaderValues.setModelMatrix(renderBufferMat4, renderBufferInvMat4);
				}

				// Build the buffers for rendering now since we have all the info.
				buildBuffers(
						camera,
						camera.getSharedTransform().getSharedInverseWorldMatrix4(),
						particleSystem);

				final Camera.RenderingValues cameraVals = camera.getSharedRenderingValues();

				if (replacementShader != null) {
					this.material.setShader(replacementShader);
				}

				this.material.setShaderVector("viewSize", cameraVals.getSharedViewSize());
				//
				this.material.setShaderVector("nearAndFarClipPlane", cameraVals.getSharedNearAndFarClipPlanes());
				///
				this.material.setShaderVector("eyePosition", cameraVals.getSharedEyePosition());
				//
				this.material.setShaderMatrix4("ViewMatrix", cameraVals.getSharedViewMatrix());
				this.material.setShaderMatrix4("invViewMatrix", cameraVals.getSharedInvViewMatrix());
				//
				this.material.setShaderMatrix4("ProjMatrix", cameraVals.getSharedProjMatrix());
				this.material.setShaderMatrix4("invProjMatrix", cameraVals.getSharedInvProjMatrix());
				//
				this.material.setShaderMatrix4("VPMatrix", cameraVals.getSharedViewProjMatrix());
				this.material.setShaderMatrix4("invVPMatrix", cameraVals.getSharedInvViewProjMatrix());
				//
				this.material.setShaderMatrix4("ModelMatrix", shaderValues.getSharedModelMatrix());
				this.material.setShaderMatrix4("invModelMatrix", shaderValues.getSharedInvModelMatrix());
				//
				this.material.setShaderMatrix4("MVMatrix", shaderValues.getSharedModelViewMatrix());
				this.material.setShaderMatrix4("invMVMatrix", shaderValues.getSharedInvModelViewMatrix());
				//
				this.material.setShaderMatrix4("MVPMatrix", shaderValues.getSharedModelViewProjMatrix());
				this.material.setShaderMatrix4("invMVPMatrix", shaderValues.getSharedInvModelViewProjMatri());
				//
				this.material.setShaderValue("isCastingShadows", this.castingShadows);
				this.material.setShaderValue("isReceivingShadows", this.receivingShadows);
				//
				final int byteStride = (POSITION_DATA_SIZE + COLOR_DATA_SIZE + UV_DATA_SIZE + PADDING_DATA_SIZE) * Float.SIZE / 8;
				final String packedVBOName = "packedVBO";
				if (!this.material.shaderVariableExists(packedVBOName)) {
					this.material.createShaderPackedVBO(
							packedVBOName,
							packedDataBuffer.capacity() * Float.SIZE / 8,
							packedDataBuffer, Shader.VBOPerfHint.DYNAMIC_DRAW);
					this.material.identifyDataInPack(packedVBOName, "position",
							POSITION_DATA_SIZE, byteStride, 0);
					int readCursor = POSITION_DATA_SIZE;
					this.material.identifyDataInPack(packedVBOName, "color",
							COLOR_DATA_SIZE, byteStride, readCursor);
					readCursor += COLOR_DATA_SIZE;
					this.material.identifyDataInPack(packedVBOName, "uv",
							UV_DATA_SIZE, byteStride, readCursor);
					readCursor += UV_DATA_SIZE;
				} else {
					this.material.updateShaderVBO(
							packedVBOName,
							packedDataBuffer.capacity() * Float.SIZE / 8,
							packedDataBuffer);
				}
				//
				this.material.setPass(this.passNum);

				this.material.use();

				if (this.triangleEBO == null) {
					this.triangleEBO = new TriangleEBO();
					this.triangleEBO.createData(
							this.drawListBuffer,
							TriangleEBO.EBOPerfHint.DYNAMIC_DRAW);
				} else {
					this.triangleEBO.updateData(this.drawListBuffer);
				}
				this.triangleEBO.drawElements();

				this.material.releaseAllShaderBuffers();

				if (replacementShader != null) {
					this.material.setShader(originalShader);
				}
			}
		}
		return stillNeedsPass();
	}


	@Override
	public boolean stillNeedsPass() {
		return this.passNum + 1 < this.material.passCount();
	}

	private ByteBuffer bbAllBuffer;
	private ByteBuffer bbParticleDrawList;

	private float[] particleAllBuffer = new float[0];
	private short[] particleDrawList = new short[0];
	private final ParticleVisitor particleVisitor = new ParticleVisitor();

	private void buildBuffers(Camera camera, Matrix4 modelView, ParticleSystem particleSystem) {
		final int stride = POSITION_DATA_SIZE + COLOR_DATA_SIZE + UV_DATA_SIZE + PADDING_DATA_SIZE;
		int particleCount = particleSystem.countAliveParticles();

		int allBufferLength = 4 * particleCount * stride;
		if (this.particleAllBuffer.length < allBufferLength) {
    		this.particleAllBuffer = new float[allBufferLength];
    		this.bbAllBuffer = ByteBuffer.allocateDirect(allBufferLength * Float.SIZE / 8);
    		this.bbAllBuffer.order(ByteOrder.nativeOrder());
    		this.packedDataBuffer = this.bbAllBuffer.asFloatBuffer();
		}

		this.particleVisitor.setVisitorPrms(particleAllBuffer, stride, camera, modelView);
		particleSystem.visitEachAliveParticle(this.particleVisitor);

		this.packedDataBuffer.position(0);
    	this.packedDataBuffer.put(this.particleAllBuffer, 0, allBufferLength);
    	this.packedDataBuffer.position(0);

		// Builds the draw-list buffer.
    	//
    	// 6 = 6 indexes for each particle.
    	final int particleDrawListLength = 6 * particleCount;
    	if (this.particleDrawList.length < particleDrawListLength) {
    		this.particleDrawList = new short[particleDrawListLength];
			this.bbParticleDrawList = ByteBuffer.allocateDirect(particleDrawListLength * Short.SIZE / 8);
			this.bbParticleDrawList.order(ByteOrder.nativeOrder());
			this.drawListBuffer = this.bbParticleDrawList.asShortBuffer();
    	}
    	for (short i = 0, vertexIndex = 0; i < particleDrawListLength; i += 6, vertexIndex += 4) {
    		// Makes a draw list such as: 0, 1, 2, 0, 2, 3.
    		particleDrawList[i    ] = vertexIndex;
    		particleDrawList[i + 1] = (short) (vertexIndex + 1);
    		particleDrawList[i + 2] = (short) (vertexIndex + 2);
    		particleDrawList[i + 3] = vertexIndex;
    		particleDrawList[i + 4] = (short) (vertexIndex + 2);
    		particleDrawList[i + 5] = (short) (vertexIndex + 3);
    	}
    	this.drawListBuffer.put(particleDrawList, 0, particleDrawListLength);
    	this.drawListBuffer.position(0);
	}

	private static final class ParticleVisitor implements ParticleSystem.IParticleVisitor {
		private int nbVisitedParticles;
		private float[] particleAllBuffer;
		private int stride;
		private Camera camera;
		private Matrix4 modelView;

		/**
		 * Sets all the parameters required to fill <code>particleAllBufferResult</code>
		 * when the method <code>visitParticle</code> will be called right after.
		 *
		 * @param particleAllBufferResult
		 * @param stride
		 * @param camera
		 * @param modelView
		 */
		private void setVisitorPrms(float[] particleAllBufferResult, int stride, Camera camera, Matrix4 modelView) {
			this.nbVisitedParticles = 0;
			this.particleAllBuffer = particleAllBufferResult;
			this.stride = stride;
			this.camera = camera;
			this.modelView = modelView;
		}

		@Override
		public void visitParticle(Particle particle) {
			final int dstStart = this.nbVisitedParticles * 4 * stride;
			getQuadParticleVertices(
					particleAllBuffer, dstStart, this.stride, this.camera, this.modelView, particle);
			getQuadParticleColors(
					particleAllBuffer, dstStart + POSITION_DATA_SIZE, this.stride, particle);
			getQuadParticleUV(
					particleAllBuffer, dstStart + POSITION_DATA_SIZE + COLOR_DATA_SIZE, this.stride, particle);
			this.nbVisitedParticles++;
		}

		private final Vector3[] rotatedParticle = {
				Vector3.getNewZeroVec(),
				Vector3.getNewZeroVec(),
				Vector3.getNewZeroVec(),
				Vector3.getNewZeroVec()
		};
		private final Vector3 particleCenter = Vector3.getNewZeroVec();
		private final Quaternion qVertexParticle = Quaternion.getNewIdQuat();

		private void getQuadParticleVertices(
				float[] particleBillboards, int offset, int stride,
				Camera camera, Matrix4 modelView, Particle particle) {
			// Gets the particle position in camera space
			modelView.transformPoint(
					this.particleCenter, particle.getSharedPosition());

			/*
			 * p0 +---+ p3
			 *    |   |
			 * p1 +---+ p2
			 */
			final float halfSize = particle.getSize() / 2.0f;

			// Builds the particle billboard with its center at (0, 0, 0).
			this.rotatedParticle[0].setValues(-halfSize,  halfSize, 0.0f);
			this.rotatedParticle[1].setValues(-halfSize, -halfSize, 0.0f);
			this.rotatedParticle[2].setValues( halfSize, -halfSize, 0.0f);
			this.rotatedParticle[3].setValues( halfSize,  halfSize, 0.0f);
			// Rotates the particle.
			this.qVertexParticle.setValues(particle.getRotation(), Vector3.BACKWARD);
			for (int i = 0; i < 4; i++) {
				this.qVertexParticle.rotateDirection(
						this.rotatedParticle[i], this.rotatedParticle[i]);
			}
			// Moves the particle to its position (reminder: we work in the camera space).
			for (int i = 0; i < 4; i++) {
				this.rotatedParticle[i].addSelf(this.particleCenter);
			}
			// Outputs the vertex values.
			for (int i = 0; i < 4; i++) {
				final int start = offset + i * stride;
				particleBillboards[start    ] = this.rotatedParticle[i].getX();
				particleBillboards[start + 1] = this.rotatedParticle[i].getY();
				particleBillboards[start + 2] = this.rotatedParticle[i].getZ();
				particleBillboards[start + 3] = 1.0f;
			}
		}

		private void getQuadParticleColors(
				float[] particleBillboards, int offset, int stride, Particle particle) {
			final float[] colorComponents = particle.getSharedColor().getSharedComponents();
			// Assigns the same color to the 4 vertices.
			for (int i = 0; i < 4; i++) {
				System.arraycopy(colorComponents, 0, particleBillboards, offset + i * stride, 4);
			}
		}

		private void getQuadParticleUV(
				float[] particleBillboards, int offset, int stride, Particle particle) {
			int dstStart = offset;
			// p0
			this.particleAllBuffer[dstStart    ] = 0.0f;
			this.particleAllBuffer[dstStart + 1] = 1.0f;
			dstStart += stride;
			// p1
			this.particleAllBuffer[dstStart    ] = 0.0f;
			this.particleAllBuffer[dstStart + 1] = 0.0f;
			dstStart += stride;
			// p2
			this.particleAllBuffer[dstStart    ] = 1.0f;
			this.particleAllBuffer[dstStart + 1] = 0.0f;
			dstStart += stride;
			// p3
			this.particleAllBuffer[dstStart    ] = 1.0f;
			this.particleAllBuffer[dstStart + 1] = 1.0f;
		}
	}

	@Override
	void updateWorldBounds() {
		final ParticleSystem particleSystem = getSharedSOb().getAnyComponent(ParticleSystem.class);
		if (particleSystem != null) {
			updateWorldBounds(particleSystem.getSharedBounds());
		}
	}
}
