package com.laguerre.kan;

import java.util.Arrays;

public final class Quaternion {
	public static final Quaternion IDENTITY = getNewIdQuat();

	private final float[] mQuat = new float[4];

	public Quaternion() {
		this(0.0f, 0.0f, 0.0f, 1.0f);
	}

	public Quaternion(float x, float y, float z, float w) {
		setValues(x, y, z, w);
	}

	public Quaternion(float angleDeg, Vector3 axis) {
		setValues(angleDeg, axis);
	}

	public Quaternion(Quaternion q) {
		setValues(q);
	}

	public void setValues(float x, float y, float z, float w) {
		this.mQuat[0] = x;
		this.mQuat[1] = y;
		this.mQuat[2] = z;
		this.mQuat[3] = w;
	}

	public void setValues(float angleDeg, Vector3 axis) {
		final float angleRad = (float) Math.toRadians(angleDeg);
		final float halfAngle = angleRad / 2.0f;
		final float sinHalfAngle = (float) Math.sin(halfAngle);
		setValues(
				axis.getX() * sinHalfAngle,
				axis.getY() * sinHalfAngle,
				axis.getZ() * sinHalfAngle,
				(float) Math.cos(halfAngle));
	}

	/**
	 * Sets this {@link Quaternion} from a matrix 3x3 column-major (as in OpenGL).
	 * @param mat3 a matrix 3x3 column-major.
	 */
	public void setValuesFromMatrix3(float[] mat3) {
		final float qw = (float) (Math.sqrt(1.0 + mat3[0] + mat3[3 + 1] + mat3[2 * 3 + 2]) / 2.0);
		final float qw4 = 4.0f * qw;
		setValues(
				(mat3[    3 + 2] - mat3[2 * 3 + 1]) / qw4,
				(mat3[2 * 3    ] - mat3[        2]) / qw4,
				(mat3[        1] - mat3[    3    ]) / qw4,
				qw);
	}

	public void setValues(Quaternion q) {
		if (q == this) {
			return;
		}
		System.arraycopy(q.mQuat, 0, this.mQuat, 0, 4);
	}

	@Override
	public String toString() {
		final float[] angleThenAxis = new float[4];
		getAngleAxis(angleThenAxis);
		return "(" + angleThenAxis[0] + " deg, [" +
			angleThenAxis[1] + ", " +
			angleThenAxis[2] + ", " +
			angleThenAxis[3] + "])";
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		} else if (!(obj instanceof Quaternion)) {
			return false;
		} else {
			final Quaternion q = (Quaternion) obj;
			boolean theSame = true;
			for (int i = 0; i < 3 && theSame; i++) {
				if (!Utils.fuzzyEquals(this.mQuat[i], q.mQuat[i])) {
					theSame = false;
				}
			}
			return theSame;
		}
	}

	@Override
	public int hashCode() {
		return Arrays.hashCode(this.mQuat);
	}

	public Quaternion addSelf(Quaternion q) {
		return add(this, this, q);
	}

	public Quaternion mulSelf(float f) {
		return mul(this, f, this);
	}

	public Quaternion mulSelf(Quaternion rQ) {
		return mul(this, this, rQ);
	}

	public Quaternion conjugateSelf() {
		return conjugate(this, this);
	}

	public Quaternion inverseSelf() {
		return inverse(this, this);
	}

	public Quaternion normalizeSelf() {
		return normalize(this, this);
	}

	public static Quaternion add(Quaternion result, Quaternion q0, Quaternion q1) {
		for (int i = 0; i < 4; i++) {
			result.mQuat[i] = q0.mQuat[i] + q1.mQuat[i];
		}
		return result;
	}

	public static Quaternion mul(Quaternion result, float f, Quaternion q) {
		for (int i = 0; i < 4; i++) {
			result.mQuat[i] = f * q.mQuat[i];
		}
		return result;
	}

	public static Quaternion mul(Quaternion result, Quaternion lQ, Quaternion rQ) {
		result.setValues(
				( lQ.mQuat[0] * rQ.mQuat[3]) + (lQ.mQuat[1] * rQ.mQuat[2]) - (lQ.mQuat[2] * rQ.mQuat[1]) + (lQ.mQuat[3] * rQ.mQuat[0]),
				(-lQ.mQuat[0] * rQ.mQuat[2]) + (lQ.mQuat[1] * rQ.mQuat[3]) + (lQ.mQuat[2] * rQ.mQuat[0]) + (lQ.mQuat[3] * rQ.mQuat[1]),
				( lQ.mQuat[0] * rQ.mQuat[1]) - (lQ.mQuat[1] * rQ.mQuat[0]) + (lQ.mQuat[2] * rQ.mQuat[3]) + (lQ.mQuat[3] * rQ.mQuat[2]),
				(-lQ.mQuat[0] * rQ.mQuat[0]) - (lQ.mQuat[1] * rQ.mQuat[1]) - (lQ.mQuat[2] * rQ.mQuat[2]) + (lQ.mQuat[3] * rQ.mQuat[3]));
		return result;
	}

	public static Quaternion conjugate(Quaternion result, Quaternion q) {
		for (int i = 0; i < 3; i++) {
			result.mQuat[i] = -q.mQuat[i];
		}
		result.mQuat[3] =  q.mQuat[3];
		return result;
	}

	public static Quaternion inverse(Quaternion result, Quaternion q) {
		conjugate(result, q);
		final float invNorm = 1.0f / getSquareNorm(q);
		return mul(result, invNorm, result);
	}

	public static float getSquareNorm(Quaternion q) {
		float accu = 0f;
		for (int i = 0; i < 4; i++) {
			accu += q.mQuat[i] * q.mQuat[i];
		}
		return accu;
	}

	public static float getNorm(Quaternion q) {
		return (float) Math.sqrt(getSquareNorm(q));
	}

	public static Quaternion normalize(Quaternion result, Quaternion q) {
		final float invNorm = 1.0f / getNorm(q);
		return mul(result, invNorm, q);
	}

	public static Quaternion add(Quaternion q0, Quaternion q1) {
		final Quaternion ret = new Quaternion();
		return mul(ret, q0, q1);
	}

	public static Quaternion mul(float f, Quaternion q) {
		final Quaternion ret = new Quaternion();
		return mul(ret, f, q);
	}

	public static Quaternion mul(Quaternion lQ, Quaternion rQ) {
		final Quaternion ret = new Quaternion();
		return mul(ret, lQ, rQ);
	}

	public static Quaternion conjugate(Quaternion q) {
		final Quaternion ret = new Quaternion();
		return conjugate(ret, q);
	}

	public static Quaternion inverse(Quaternion q) {
		final Quaternion ret = new Quaternion();
		return inverse(ret, q);
	}

	public static Quaternion normalize(Quaternion q) {
		final Quaternion ret = new Quaternion();
		return normalize(ret, q);
	}

	public float getX() {
		return this.mQuat[0];
	}

	public float getY() {
		return this.mQuat[1];
	}

	public float getZ() {
		return this.mQuat[2];
	}

	public float getW() {
		return this.mQuat[3];
	}

	/**
	 * Gets the value of a component of this {@link Quaternion}.
	 * <ul>
	 *   <li>index 0 equals to getX()</li>
	 *   <li>index 1 equals to getY()</li>
	 *   <li>index 2 equals to getZ()</li>
	 *   <li>index 3 equals to getW()</li>
	 * </ul>
	 * @param i between 0 and 4 (excluded) to get x, y, z or w.
	 * @return the value of x, y, z, w.
	 */
	public float[] getComponents(float[] result) {
		System.arraycopy(this.mQuat, 0, result, 0, 4);
		return result;
	}

	public float getComponent(int i) {
		return this.mQuat[i];
	}

	/**
	 * Gets the form (angle, axis) of this {@link Quaternion}, that you should
	 * have been normalized to provide a correct result.
	 *
	 * @param angleThenAxis - angleThenAxis[0] is the angle between 0 and PI,
	 *   whilst angleThenAxis[1..3] are the coordinates of the axis.
	 */
	public void getRadianAngleAxis(float[] angleThenAxis) {
		/*
		 * angle = 2 * acos(qw)
		 * x = qx / sqrt(1-qw*qw)
		 * y = qy / sqrt(1-qw*qw)
		 * z = qz / sqrt(1-qw*qw)
		 */
		angleThenAxis[0] = 2.0f * (float) Math.acos(this.mQuat[3]);

		final float f = (float) Math.sqrt(1.0f - (this.mQuat[3] * this.mQuat[3]));
		if (f != 0.0f) {
			for (int i = 0; i < 3; i++) {
				angleThenAxis[i + 1] = this.mQuat[i] / f;
			}
		} else {
			// Use an arbitrary axis, here (0, 0, 1).
			angleThenAxis[1] = angleThenAxis[2] = 0.0f;
			angleThenAxis[3] = 1.0f;
		}
	}

	/**
	 * Gets the form (angle, axis) of this {@link Quaternion}, that you should
	 * have been normalized to provide a correct result.
	 *
	 * @param angleThenAxis - angleThenAxis[0] is the angle between 0 and 180,
	 *   whilst angleThenAxis[1..3] are the coordinates of the axis.
	 */
	public void getAngleAxis(float[] angleThenAxis) {
		getRadianAngleAxis(angleThenAxis);
		angleThenAxis[0] = (float) Math.toDegrees(angleThenAxis[0]);
	}

	/**
	 * Used by <code>rotateDirection</code> to store the {@link Vector3} that as
	 * a {@link Quaternion} to rotate it. This field avoids many memory allocations.
	 * @see Quaternion.rotateDirection
	 */
	private static final Quaternion tmpRotateVecAsQuat = new Quaternion();
	/**
	 * Used by <code>rotateDirection</code> to store this {@link Quaternion} in
	 * order to apply operations on it. This field avoids many memory allocations.
	 * @see Quaternion.rotateDirection
	 */
	private static final Quaternion tmpRotateThis = new Quaternion();

	public Vector3 rotateDirection(Vector3 result, Vector3 v) {
		synchronized (tmpRotateVecAsQuat) {
			tmpRotateVecAsQuat.setValues(v.getX(), v.getY(), v.getZ(), 0.0f);
			// vq' = this * vq * inverse(this)
			// but since we assume that the quaternion is normalized:
			// vq' = this * vq * conjugate(this)
			Quaternion.conjugate(tmpRotateThis, this);
			Quaternion.mul(tmpRotateVecAsQuat, tmpRotateVecAsQuat, tmpRotateThis);
			Quaternion.mul(tmpRotateVecAsQuat, this, tmpRotateVecAsQuat);
			result.setValues(
					tmpRotateVecAsQuat.mQuat[0],
					tmpRotateVecAsQuat.mQuat[1],
					tmpRotateVecAsQuat.mQuat[2]);
		}
		return result;
	}

	public Vector3 rotateDirection(Vector3 v) {
		final Vector3 ret = new Vector3();
		return rotateDirection(ret, v);
	}

	public static Quaternion getNewIdQuat() {
		return new Quaternion(0.0f, 0.0f, 0.0f, 1.0f);
	}
}
