package com.laguerre.kan;

import java.nio.ShortBuffer;

import android.opengl.GLES20;

public abstract class Renderer extends Component {
	/**
	 * Stores indices as an element buffer object and makes it possible to draw
	 * triangles associated with these indices.
	 */
	static class TriangleEBO {
		public static enum EBOPerfHint {
			STATIC_DRAW,
			STREAM_DRAW,
			DYNAMIC_DRAW
		}

		public final int eboId;
		private boolean beingNew = true;
		private int bufferSize = -1;
		private int nbIndices = -1;
		private EBOPerfHint perfHint;

		public TriangleEBO() {
			final int[] tmpId = { -1 };
			GLES20.glGenBuffers(1, tmpId, 0);
			this.eboId = tmpId[0];
		}

		public boolean isNew() {
			return this.beingNew;
		}

		public void createData(ShortBuffer triangleBuffer, EBOPerfHint perfHint) {
			this.beingNew = false;
			this.perfHint = perfHint;
			this.nbIndices = triangleBuffer.capacity();
			GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, this.eboId);
			int drawMode;
			switch (this.perfHint) {
			case STREAM_DRAW:
				drawMode = GLES20.GL_STREAM_DRAW;
				break;
			case DYNAMIC_DRAW:
				drawMode = GLES20.GL_DYNAMIC_DRAW;
				break;
			case STATIC_DRAW:
			default:
				drawMode = GLES20.GL_STATIC_DRAW;
				break;
			}
			this.bufferSize = triangleBuffer.capacity() * Short.SIZE / 8;
			GLES20.glBufferData(GLES20.GL_ELEMENT_ARRAY_BUFFER,
					this.bufferSize, triangleBuffer, drawMode);
			GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, 0);
		}

		public void updateData(ShortBuffer triangleBuffer) {
			if (this.perfHint == EBOPerfHint.STATIC_DRAW) {
				return;
			}
			final int newByteSize = triangleBuffer.capacity() * Short.SIZE / 8;
			if (newByteSize != this.bufferSize) {
				createData(triangleBuffer, this.perfHint);
			} else {
				GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, this.eboId);
				GLES20.glBufferSubData(GLES20.GL_ELEMENT_ARRAY_BUFFER, 0,
						this.bufferSize, triangleBuffer);
				GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, 0);
			}
		}

		public void drawElements() {
			GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, this.eboId);
			AndroidGL20Compact.getInstance().glDrawElements(GLES20.GL_TRIANGLES,
					this.nbIndices, GLES20.GL_UNSIGNED_SHORT, 0);
//			GLES20.glDrawElements(GLES20.GL_TRIANGLES,
//					this.nbIndices, GLES20.GL_UNSIGNED_SHORT, 0);
			GLES20.glBindBuffer(GLES20.GL_ELEMENT_ARRAY_BUFFER, 0);
		}

		public void freeResources() {
			final int[] tmpId = { this.eboId };
			GLES20.glDeleteBuffers(1, tmpId, 0);
		}
	}

	protected final Bounds worldBounds = new Bounds(Vector3.ZERO, Vector3.ZERO);
	protected boolean castingShadows = true;
	protected boolean receivingShadows = true;

	/**
	 * Sets the {@link Material} of the sub-mesh with a given Id.
	 * @param mat the {@link Material} to associate to a sub-mesh.
	 * @param subMeshId the Id of the sub-mesh that will receive a {@link Material}.
	 */
	public abstract void setMaterial(Material mat, int subMeshId);

	/**
	 * Gets the {@link Material} of the sub-mesh with a given Id.
	 * @param subMeshId the Id of the sub-mesh to treat.
	 * @return the {@link Material} or <code>null</code>.
	 */
	public abstract Material getSharedMaterial(int subMeshId);

	/**
	 * Gets the Id of the sub-mesh that uses the {@link Material} named
	 * <code>materialName</code>.
	 * @param materialName the name of the {@link Material} to be found.
	 * @return the Id or <code>-1</code> if nothing has been found.
	 * @see Material.getMaterialName
	 */
	public abstract int findSubMeshIdOfMaterial(String materialName);

	/**
	 * Gets the first found {@link Material} that is named <code>materialName</code>.
	 * @param materialName the name of the {@link Material} to be found.
	 * @return a reference to the {@link Material} or <code>null</code>.
	 * @see Material.getMaterialName
	 */
	public abstract Material findMaterial(String materialName);

	/**
	 * Sets whether this {@link Renderer} casts shadows.
	 * @param castingShadows <code>true</code> if this {@link Renderer}
	 *   casts shadows, <code>false</code> otherwise.
	 */
	public final void setCastShadows(boolean castingShadows) {
		this.castingShadows = castingShadows;
	}

	/**
	 * Tells whether this {@link Renderer} is casting shadows.
	 * @return <code>true</code> if this {@link Renderer} casts shadows,
	 *   <code>false</code> otherwise.
	 */
	public final boolean isCastingShadows() {
		return this.castingShadows;
	}

	/**
	 * Sets whether this {@link Renderer} receives shadows.
	 * @param receivingShadows <code>true</code> if this {@link Renderer}
	 *   receives shadows, <code>false</code> otherwise.
	 */
	public final void setReceiveShadows(boolean receivingShadows) {
		this.receivingShadows = receivingShadows;
	}

	/**
	 * Tells whether this {@link Renderer} is receiving shadows.
	 * @return <code>true</code> if this {@link Renderer} receives shadows,
	 *   <code>false</code> otherwise.
	 */
	public final boolean isReceivingShadows() {
		return this.receivingShadows;
	}

	/**
	 * Sets the rendering pass to the first one. The method <code>getCurrentPass</code>
	 * will return <code>0</code> right after its call and before a call to
	 * <code>renderOnePass</code>.
	 * @see Renderer.getCurrentPass
	 * @see Renderer.renderOnePass
	 */
	public abstract void setToFirstPass();

	/**
	 * Gets the number of the current pass.
	 * @return the number of the current pass. <code>0</code> for the first pass.
	 */
	public abstract int getCurrentPass();

	/**
	 * Renders one pass, only if at least one is needed. Rendering will draw
	 * triangles of the involved geometry.
	 * @param camera the {@link Camera} that is rendering.
	 * @return <code>true</code> if at least one pass is still needed.
	 * @see Renderer.stillNeedsPass
	 */
	public abstract boolean renderOnePass(Camera camera);

	/**
	 * Tells whether at least one pass is needed by at least one {@link Material}
	 * of the associated geometry.
	 * @return <code>true</code> if at least one pass is still needed,
	 *   <code>false</code> otherwise.
	 */
	public abstract boolean stillNeedsPass();

	/**
	 * Gets the {@link Bounds} (i.e. AABB) of the geometry being rendered in
	 * world coordinates.
	 * @return the {@link Bounds} as AABB.
	 */
	public final Bounds getSharedWorldBounds() {
		return this.worldBounds;
	}

	abstract void updateWorldBounds();

	/**
	 * Stores the 8 vertices of an AABB. Used by
	 * <code>updateWorldBounds(Bounds localBounds)</code> only to avoid many
	 * memory allocations.
	 */
	private static final Vector3[] points_worldBounds = {
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec(),
			Vector3.getNewZeroVec()
	};
	/**
	 * Stores the minimum values from the 8 vertex coordinates of an AABB.
	 * Used by <code>updateWorldBounds(Bounds localBounds)</code> only to avoid
	 * many memory allocations.
	 */
	private static final Vector3 min_worldBounds = Vector3.getNewZeroVec();
	/**
	 * Stores the maximum values from the 8 vertex coordinates of an AABB.
	 * Used by <code>updateWorldBounds(Bounds localBounds)</code> only to avoid
	 * many memory allocations.
	 */
	private static final Vector3 max_worldBounds = Vector3.getNewZeroVec();

	protected final void updateWorldBounds(Bounds localBounds) {
		synchronized (points_worldBounds) {
			localBounds.getHeightVertices(points_worldBounds);
			final Transform t = getSharedTransform();
			for (int i = 0; i < points_worldBounds.length; i++) {
				t.transformWorldPoint(
						points_worldBounds[i], points_worldBounds[i]);
			}

			float minX = Float.MAX_VALUE, minY = Float.MAX_VALUE, minZ = Float.MAX_VALUE;
			float maxX = Float.MIN_VALUE, maxY = Float.MIN_VALUE, maxZ = Float.MIN_VALUE;
			for (int i = 0; i < points_worldBounds.length; i++) {
				final float x = points_worldBounds[i].getX();
				final float y = points_worldBounds[i].getY();
				final float z = points_worldBounds[i].getZ();
				if (x < minX) {
					minX = x;
				}
				if (y < minY) {
					minY = y;
				}
				if (z < minZ) {
					minZ = z;
				}
				if (x > maxX) {
					maxX = x;
				}
				if (y > maxY) {
					maxY = y;
				}
				if (z > maxZ) {
					maxZ = z;
				}
			}
			min_worldBounds.setValues(minX, minY, minZ);
			max_worldBounds.setValues(maxX, maxY, maxZ);
			this.worldBounds.setFromMinAndMax(min_worldBounds, max_worldBounds);
		}
	}
}
