package com.laguerre.kan;

import java.util.ArrayList;
import java.util.HashSet;

import android.util.SparseArray;

public final class SceneConductor {
	public interface ISurfaceChangedListener {
		public void onSurfaceChanged(int width, int height);
	}

	public static interface ISceneObjectVisitor {
		public void visitSceneObject(SceneObject sob);
	}

	public static interface IRenderQueueVisitor {
		public void visit(SceneObject sob, int queue);
	}

	private static SceneConductor instance = null;
	private final HashSet<ISurfaceChangedListener> surfaceChangedListeners =
			new HashSet<ISurfaceChangedListener>();
	private final SparseArray<SceneObject> sobs = new SparseArray<SceneObject>();
	private final ArrayList<SceneObject> camsToRender = new ArrayList<SceneObject>();
	private final SparseArray<ArrayList<SceneObject>> queues = new SparseArray<ArrayList<SceneObject>>();

	public static SceneConductor getInstance() {
		if (instance == null) {
			instance = new SceneConductor();
		}
		return instance;
	}
	public static void resetInstance() {
		instance = null;
	}

	private SceneConductor() {
		this.queues.put(Shader.RenderQueueType.BACKGROUND.index(), new ArrayList<SceneObject>());
		this.queues.put(Shader.RenderQueueType.GEOMETRY.index(), new ArrayList<SceneObject>());
		this.queues.put(Shader.RenderQueueType.TRANSPARENT.index(), new ArrayList<SceneObject>());
	}

	public void onRenderSurfaceChanged(int screenWidth, int screenHeight) {
		for (ISurfaceChangedListener listener : this.surfaceChangedListeners) {
			listener.onSurfaceChanged(screenWidth, screenHeight);
		}
	}

	public void addSurfaceChangedListener(ISurfaceChangedListener listener) {
		this.surfaceChangedListeners.add(listener);
	}

	public void removeSurfaceChangedListener(ISurfaceChangedListener listener) {
		this.surfaceChangedListeners.remove(listener);
	}

	public SceneObject addSceneObject() {
		return addSceneObject((String) null);
	}

	public SceneObject addSceneObject(String sobName) {
		final SceneObject sob = new SceneObject();
		sob.setName(sobName);
		this.sobs.put(sob.getId(), sob);
		return sob;
	}

	public void destroySceneObject(SceneObject sob) {
		if (sob != null) {
			this.sobs.remove(sob.getId());
			sob.destroy();
		}
	}

	public void destroySceneObject(int sobId) {
		destroySceneObject(this.sobs.get(sobId));
	}

	public void visitEachSceneObject(ISceneObjectVisitor visitor) {
		for (int i = 0, sobsSize = this.sobs.size(); i < sobsSize; i++) {
			final SceneObject sob = this.sobs.valueAt(i);
			visitor.visitSceneObject(sob);
		}
	}

	public int sceneObjectCount() {
		return this.sobs.size();
	}

	public void addSceneObjectToRenderQueue(SceneObject sob, int queue) {
		ArrayList<SceneObject> objQueue = this.queues.get(queue);
		if (objQueue == null) {
			objQueue = new ArrayList<SceneObject>();
			this.queues.put(queue, objQueue);
		}
		if (!objQueue.contains(sob)) {
			objQueue.add(sob);
		}
	}

	public void removeSceneObjectFromQueue(SceneObject sob, int queue) {
		ArrayList<SceneObject> objQueue = this.queues.get(queue);
		if (objQueue != null) {
			objQueue.remove(sob);
		}
	}

	public void removeSceneObjectFromQueue(SceneObject sob) {
		for (int i = 0, iEnd = this.queues.size(); i < iEnd; i++) {
			removeSceneObjectFromQueue(sob, i);
		}
	}

	public int countRenderQueues() {
		return this.queues.size();
	}

	/**
	 * Gets the index associated to a queue.
	 * @param queue
	 * @return the index or a negative value if not found.
	 */
	public int getRenderQueueIndex(int queue) {
		return this.queues.indexOfKey(queue);
	}

	/**
	 * Gets the queue associated to an index.
	 * @param queueIndex
	 * @return the queue or a negative value if not found.
	 */
	public int getQueueAtIndex(int queueIndex) {
		int ret = -1;
		if (queueIndex > 0 && queueIndex < this.queues.size()) {
			ret = this.queues.keyAt(queueIndex);
		}
		return ret;
	}

	/**
	 * Visits all the {@link SceneObject} associated to a queue index. The method
	 * <code>visit</code> of {@link IRenderQueueVisitor} will be called on each
	 * scene object found.
	 * @param queueIndex index of queue (i.e. an integer between 0 and <code>countRenderQueues</code>.
	 * @param visitor
	 * @see countRenderQueues
	 * @see getRenderQueueIndex
	 * @see getQueueAtIndex
	 */
	public void visitRenderQueue(int queueIndex, IRenderQueueVisitor visitor) {
		final ArrayList<SceneObject> objQueue = this.queues.valueAt(queueIndex);
		if (objQueue != null) {
			final int queue = this.queues.keyAt(queueIndex);
			for (int sob = 0, sobEnd = objQueue.size(); sob < sobEnd; sob++) {
				visitor.visit(objQueue.get(sob), queue);
			}
		}
	}

	void addCamera(SceneObject camSob) {
		if (!this.camsToRender.contains(camSob)) {
			this.camsToRender.add(camSob);
		}
	}

	void removeCamera(SceneObject camSob) {
		this.camsToRender.remove(camSob);
	}

	public void drawFrame() {
		Time.getInstance().resetTimeFrameStarted();

		// First update states,
		for (int i = 0, sobsSize = this.sobs.size(); i < sobsSize; i++) {
			final SceneObject sob = this.sobs.valueAt(i);
			sob.doLifeStep();
		}

		final Camera currentCam = Camera.getCurrent();

		// Ask lights to make their shadow map (if they should).
		SceneObject brightestLight =
				Lights.getInstance().getBrightestDirectionalLight(currentCam);
		if (brightestLight != null) {
			brightestLight.getSharedLight().renderToShadowDepthMap();
		}

		// Then draw objects.
		//
		// First all the cameras that render to textures in the scene.
		for (int i = 0, iEnd = this.camsToRender.size(); i < iEnd; i++) {
			final Camera sharedCam = this.camsToRender.get(i).getSharedCamera();
			if (currentCam != sharedCam && sharedCam.isEnabled()) {
				sharedCam.render();
			}
		}
		// Finally the main camera.
		currentCam.render();

		Time.getInstance().captureFrameElapsedTime();
		Time.getInstance().incrFrameCount();
	}
}
