package com.laguerre.kan;

import android.util.SparseArray;

public final class SceneObject extends BaseObject {
	private boolean enabled;
	private int layer;
	private String tag;
	private final Transform transform = new Transform();
	private final SparseArray<Component> components = new SparseArray<Component>();
	private final SparseArray<Behaviour> behaviours = new SparseArray<Behaviour>();

	// Shortcuts
	private Renderer renderer = null;
	private Camera camera = null;
	private Light light = null;
	private ParticleSystem particleSystem = null;

	SceneObject() {
		this.enabled = true;
		this.layer = 1 << 0;
	}

	public boolean isEnabled() {
		return this.enabled;
	}

	public void setEnabled(boolean enabled) {
		this.enabled = enabled;
		final int behaviourSize = this.behaviours.size();
		for (int i = 0; i < behaviourSize; i++) {
			final Behaviour behaviour = this.behaviours.valueAt(i);
			behaviour.setEnabled(enabled);
		}
	}

	public void setTag(String tag) {
		this.tag = tag;
	}

	public String getTag() {
		return this.tag;
	}

	public void setLayer(int layer) {
		this.layer = layer;
	}

	public int getLayer() {
		return this.layer;
	}

	public Transform getSharedTransform() {
		return this.transform;
	}

	public Renderer getSharedRenderer() {
		return this.renderer;
	}

	public Camera getSharedCamera() {
		return this.camera;
	}

	public Light getSharedLight() {
		return this.light;
	}

	public final ParticleSystem getSharedParticleSystem() {
		return this.particleSystem;
	}

	public <T extends Component> T getAnyComponent(Class<T> type) {
		Component foundComponent = null;
		final int componentSize = this.components.size();
		for (int i = 0; i < componentSize && foundComponent == null; i++) {
			final Component component = this.components.valueAt(i);
			if (type.isAssignableFrom(component.getClass())) {
				foundComponent = component;
			}
		}
		return type.cast(foundComponent);
	}

	/**
	 * Updates the states of the {@link Behaviour} objects of this {@link SceneObject}
	 * if <code>isEnabled</code> returned <code>true</code>. Right before
	 * behaviors, the states of {@link Component} objects that may require
	 * updated states along time are treated, such as {@link ParticleSystem}.
	 * For these objects the method called is class dependent, which means that
	 * different names or architecture designs could be used (e.g. <code>doLifeStep</code>
	 * is called for {@link ParticleSystem}).
	 *
	 * @see SceneObject.isEnabled
	 */
	void doLifeStep() {
		if (!isEnabled()) {
			return;
		}
		if (this.particleSystem != null) {
			this.particleSystem.doLifeStep();
		}
		final int behaviourSize = this.behaviours.size();
		for (int i = 0; i < behaviourSize; i++) {
			final Behaviour behaviour = this.behaviours.valueAt(i);
			behaviour.doLifeStep();
		}
	}

	/**
	 * Called by a {@link Camera} just before it render one pass to dispatch
	 * this event to the {@link Behaviour} that is associated with the {@link Camera}.
	 * @param passNum
	 */
	void onBeforeRenderingCameraPass(int passNum) {
		final int behaviourSize = this.behaviours.size();
		for (int i = 0; i < behaviourSize; i++) {
			final Behaviour behaviour = this.behaviours.valueAt(i);
			if (behaviour.isEnabled()) {
				behaviour.onBeforeRenderingCameraPass(passNum);
			}
		}
	}

	/**
	 * Called before one pass that will apply an image effect, to dispatch this
	 * event to the {@link Behaviour} that is associated with the {@link Camera}.
	 *
	 * The first image that will be provided is the result of the render passes
	 * of the current {@link Camera}.
	 *
	 * @param passNum
	 */
	public void onBeforeRenderingImageEffectsPass(int passNum) {
		final int behaviourSize = this.behaviours.size();
		for (int i = 0; i < behaviourSize; i++) {
			final Behaviour behaviour = this.behaviours.valueAt(i);
			if (behaviour.isEnabled()) {
				behaviour.onBeforeRenderingImageEffectsPass(passNum);
			}
		}
	}

	final void destroy() {
		final int behaviourSize = this.behaviours.size();
		for (int i = 0; i < behaviourSize; i++) {
			final Behaviour behaviour = this.behaviours.valueAt(i);
			behaviour.destroy();
		}
	}

	public <T extends Component> T addComponent(Class<T> componentClass) {
		return addComponent(componentClass, null);
	}

	public <T extends Component> T addComponent(Class<T> componentClass, String name) {
		T ret = null;
		try {
			ret = componentClass.newInstance();
			ret.setParentSceneObject(this);
			ret.setName(name);
			storeAddedComponent(ret);
		} catch (IllegalArgumentException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}
		return ret;
	}

	private <T extends Component> void storeAddedComponent(T component) {
		this.components.put(component.getId(), component);
		// if possible, store it also as SObBehaviour.
		if (component instanceof Behaviour) {
			final Behaviour behaviour = (Behaviour) component;
			this.behaviours.put(behaviour.getId(), behaviour);
		}
		// Is this component a shortcut that we handle?
		if (component instanceof Renderer) {
			if (this.renderer != null) {
				throw new DuplicatedComponent("Renderer already exists for this SOb.");
			}
			this.renderer = (Renderer) component;
		} else if (component instanceof Camera) {
			if (this.camera != null) {
				throw new DuplicatedComponent("Camera already exists for this SOb.");
			}
			this.camera = (Camera) component;
			// Make the camera "working".
			SceneConductor.getInstance().addCamera(this);
		} else if (component instanceof Light) {
			if (this.light != null) {
				throw new DuplicatedComponent("Light already exists for this SOb.");
			}
			this.light = (Light) component;
			Lights.getInstance().addLight(this);
		} else if (component instanceof ParticleSystem) {
			if (this.particleSystem != null) {
				throw new DuplicatedComponent("ParticleSystem already exists for this SOb.");
			}
			this.particleSystem = (ParticleSystem) component;
		}
	}

	public <T extends Component> void removeComponent(T component) {
		this.components.remove(component.getId());
		if (component instanceof Behaviour) {
			this.behaviours.remove(component.getId());
		}
		// Is this component a shortcut that we handle?
		if (component instanceof Renderer) {
			this.renderer = null;
		} else if (component instanceof Camera) {
			this.camera = null;
			// Make the camera "not working".
			SceneConductor.getInstance().removeCamera(this);
		} else if (component instanceof Light) {
			Lights.getInstance().removeLight(this);
			this.light = null;
		} else if (component instanceof ParticleSystem) {
			this.particleSystem = null;
		}
	}

	public static class DuplicatedComponent extends RuntimeException {
		private static final long serialVersionUID = 7746453127776700693L;

		public DuplicatedComponent(String detailMessage) {
			super(detailMessage);
		}
	}
}
