package com.laguerre.kan;

import java.util.ArrayList;

import android.opengl.Matrix;

public final class Transform {
	private Transform parent = null;
	private ArrayList<Transform> children = new ArrayList<Transform>();

	private final Vector3 worldTranslation = Vector3.getNewZeroVec();
	private final Quaternion worldRotation = Quaternion.getNewIdQuat();
	private final Vector3 worldDerivedScale= Vector3.getNewOneVec();

	private final Vector3 localTranslation = Vector3.getNewZeroVec();
	private final Quaternion localRotation = Quaternion.getNewIdQuat();
	private final Vector3 localScale= Vector3.getNewOneVec();

	private final Matrix4 worldMatrix = Matrix4.getNewIdentity();
	private boolean worldDirty = false;
	private final Matrix4 localMatrix = Matrix4.getNewIdentity();
	private boolean localDirty = false;
	private final Matrix4 cachedInverseWorldMatrix = Matrix4.getNewIdentity();
	private boolean inverseWorldDirty = false;
	private final Matrix4 cachedInverseLocalMatrix = Matrix4.getNewIdentity();
	private boolean inverseLocalDirty = false;

	/**
	 * Constant to compare against identity in order to avoid the creation of
	 * many intermediate objects.
	 * <br />
	 * Since Java cannot ensure constantness of the returned object, please do
	 * <b>not</b> modify it.
	 */
	public static final Transform IDENTITY = getNewIdentity();

	public static Transform getNewIdentity() {
		return new Transform();
	}

	public Transform() {
		this(Vector3.ZERO);
	}

	public Transform(Vector3 translation) {
		this(translation, Quaternion.IDENTITY);
	}

	public Transform(Quaternion rotation) {
		this(Vector3.ZERO, rotation);
	}

	public Transform(Vector3 translation, Quaternion rotation) {
		this(translation, rotation, Vector3.ONE);
	}

	public Transform(Vector3 translation, Quaternion rotation, Vector3 scale) {
		setLocalValues(translation, rotation, scale);
	}

	public Transform(Transform t) {
		setValues(t);
	}

	public void setValues(Transform t) {
		if (t == this) {
			return;
		}
		this.parent = t.parent;
		// Replace children.
		this.children.clear();
		if (!t.children.isEmpty()) {
			this.children.addAll(t.children);
		}
		// Local transform
		this.localTranslation.setValues(t.localTranslation);
		this.localRotation.setValues(t.localRotation);
		this.localScale.setValues(t.localScale);
		// World transform
		this.worldTranslation.setValues(t.worldTranslation);
		this.worldRotation.setValues(t.worldRotation);
		this.worldDerivedScale.setValues(t.worldDerivedScale);
		// Matrixes and dirtiness
		this.localMatrix.setValues(t.localMatrix);
		this.localDirty = t.localDirty;
		this.worldMatrix.setValues(t.worldMatrix);
		this.worldDirty = t.worldDirty;
		this.cachedInverseLocalMatrix.setValues(t.cachedInverseLocalMatrix);
		this.inverseLocalDirty = t.inverseLocalDirty;
		this.cachedInverseWorldMatrix.setValues(t.cachedInverseWorldMatrix);
		this.inverseWorldDirty = t.inverseWorldDirty;
	}

	public void setWorldValues(Vector3 translation, Quaternion rotation) {
		setWorldValues(this.parent, translation, rotation);
	}

	public void setLocalValues(Vector3 translation, Quaternion rotation, Vector3 scale) {
		setLocalValues(this.parent, translation, rotation, scale);
	}

	public void setWorldValues(Transform parent, Vector3 translation, Quaternion rotation) {
		setParent(parent);
		this.worldDerivedScale.setValues(Vector3.ONE);
		this.worldRotation.setValues(rotation);
		this.worldTranslation.setValues(translation);
		this.worldDirty = true;
		this.inverseWorldDirty = true;
		updateLocalTransform();
	}

	public void setLocalValues(Transform parent, Vector3 translation, Quaternion rotation, Vector3 scale) {
		setParent(parent);
		this.localScale.setValues(scale);
		this.localRotation.setValues(rotation);
		this.localTranslation.setValues(translation);
		this.localDirty = true;
		this.inverseLocalDirty = true;
		updateWorldTransformAndChildren();
	}

	@Override
	public String toString() {
		return "{lT: " + this.localTranslation +
				", lR: " + this.localRotation +
				", lS: " + this.localScale +
				"}, {WT: " + this.worldTranslation +
				", WR: " + this.worldRotation +
				", WS: " + this.worldDerivedScale +
				"}}";
	}

	public boolean equalsIgnoringParent(Object obj) {
		return objectEquals(obj, true);
	}

	@Override
	public boolean equals(Object obj) {
		return objectEquals(obj, false);
	}

	private boolean objectEquals(Object obj, boolean ignoreParent) {
		if (this == obj) {
			return true;
		} else if (!(obj instanceof Transform)) {
			return false;
		} else {
			final Transform t = (Transform) obj;
			final boolean parentsEqual = (ignoreParent ? true : this.parent == t.parent);
			return parentsEqual &&
					// Since world and local representations are synchronized,
					// we compare only the world representation.
					this.worldDerivedScale.equals(t.worldDerivedScale) &&
					this.worldRotation.equals(t.worldRotation) &&
					this.worldTranslation.equals(t.worldTranslation);
		}
	}

	@Override
	public int hashCode() {
		// Ignores "parent" to avoid a up-traversing of all the tree in order to
		// get hashcodes until the root Transform.
		// So we may end up with two Transform that are different (according to
		// equals()) but with the same hashcode, which is fine:
		// http://docs.oracle.com/javase/6/docs/api/java/lang/Object.html#hashCode%28%29
		return this.worldDerivedScale.hashCode() ^ this.worldRotation.hashCode() ^ this.worldTranslation.hashCode();
	}

	public Matrix4 getWorldMatrix4(Matrix4 result) {
		updateWorldMatrixIfDirty();
		result.setValues(this.worldMatrix);
		return result;
	}

	public Matrix4 getWorldMatrix4() {
		final Matrix4 ret = new Matrix4();
		return getWorldMatrix4(ret);
	}

	public float[] getWorldMatrix4(float[] result) {
		updateWorldMatrixIfDirty();
		System.arraycopy(this.worldMatrix.getSharedComponents(), 0, result, 0, 16);
		return result;
	}

	private void updateWorldMatrixIfDirty() {
		if (this.worldDirty) {
			updateLocalMatrixIfDirty();
			if (hasParent()) {
				this.parent.updateWorldMatrixIfDirty();
				Matrix4.combine(this.worldMatrix, this.parent.worldMatrix, this.localMatrix);
			} else {
				this.worldMatrix.setValues(this.localMatrix);
			}
			this.worldDirty = false;
		}
	}

	public Matrix4 getLocalMatrix4(Matrix4 result) {
		updateLocalMatrixIfDirty();
		result.setValues(this.localMatrix);
		return result;
	}

	public Matrix4 getLocalMatrix4() {
		final Matrix4 ret = new Matrix4();
		return getLocalMatrix4(ret);
	}

	public float[] getLocalMatrix4(float[] result) {
		updateLocalMatrixIfDirty();
		System.arraycopy(this.localMatrix.getSharedComponents(), 0, result, 0, 16);
		return result;
	}

	private void updateLocalMatrixIfDirty() {
		if (this.localDirty) {
			this.localMatrix.setToIdentity();
			this.localMatrix.scale(this.localScale);
			this.localMatrix.rotate(this.localRotation);
			this.localMatrix.translate(this.localTranslation);

			this.localDirty = false;
		}
	}

	public boolean hasParent() {
		return this.parent != null;
	}

	/**
	 * Sets a parent to this {@link Transform}. Note that cyclic graphs are not
	 * supported.
	 *
	 * @param parent - The parent of this {@link Transform}, which cannot be itself.
	 */
	public void setParent(Transform parent) {
		if (this.parent != parent) {
			// Detach from previous parent.
			if (this.parent != null) {
				this.parent.detachChild(this);
			}
			// Set and attach to the new parent
			this.parent = parent;
			if (this.parent != null) {
				this.parent.attachChild(this);
			}
		}
	}

	/**
	 * Gets the immediate parent of this {@link Transform}.
	 *
	 * @return the parent {@link Transform} or <code>null</code> if none.
	 */
	public Transform getSharedParent() {
		return this.parent;
	}

	/**
	 * Gets the top-most {@link Transform} in the hierarchy. This method will
	 * never return <code>null</code> but the {@link Transform} itself.
	 *
	 * @return the top-most {@link Transform} or itself.
	 */
	public Transform getRootSharedParent() {
		Transform root = this;
		while (root.hasParent()) {
			root = root.parent;
		}
		return root;
	}

	private void attachChild(Transform t) {
		if (!this.children.contains(t)) {
			this.children.add(t);
		}
	}

	private void detachChild(Transform t) {
		this.children.remove(t);
	}

	/**
	 * Gets the immediate children of this {@link Transform}.
	 *
	 * @param result - the children.
	 * @return <code>result</code> (that contains the children).
	 */
	public Transform[] getChildren(Transform[] result) {
		this.children.toArray(result);
		return result;
	}

	/**
	 * Gets the immediate children of this {@link Transform}.
	 *
	 * @return an array with the children.
	 */
	public Transform[] getChildren() {
		return (Transform[]) this.children.toArray();
	}

	/**
	 * Tells whether a {@link Transform} is an immediate child of this {@link Transform}.
	 *
	 * @param t - The {@link Transform} to test.
	 * @return <code>true</code> if the given {@link Transform} is a child,
	 *   <code>false</code> otherwise.
	 */
	public boolean isChildOf(Transform t) {
		return t.children.contains(this);
	}

	/**
	 * Tells whether a {@link Transform} is a descendant of this {@link Transform}.
	 * It means that it may be a child through many intermediate parent {@link Transform}.
	 * Note that <code>this.isDescendantOf(this)</code> returns <code>false</code>.
	 *
	 * @param t - The {@link Transform} to test. The value <code>null</code> is
	 *   accepted since it is a {@link Transform} without any parent, but every
	 *   {@link Transform} is also descendant of <code>null</code>.
	 * @return <code>true</code> if the given {@link Transform} is a descendant,
	 *   <code>false</code> otherwise.
	 */
	public boolean isDescendantOf(Transform t) {
		Transform curr = this.parent;
		while (curr != t && curr != null) {
			curr = curr.parent;
		}
		return curr == t;
	}

	public Vector3 getWorldDerivedScale(Vector3 result) {
		result.setValues(this.worldDerivedScale);
		return result;
	}

	public Vector3 getWorldDerivedScale() {
		return new Vector3(this.worldDerivedScale);
	}

	/**
	 * Gets a direct reference to the inner <code>Vector3</code> that contains
	 * scale in order to avoid memory allocations. Please, do <b>not</b> update
	 * scale components by this way.
	 * @return A reference to the scale.
	 */
	public Vector3 getSharedWorldDerivedScale() {
		return this.worldDerivedScale;
	}

	public Quaternion getWorldRotation(Quaternion result) {
		result.setValues(this.worldRotation);
		return result;
	}

	public Quaternion getWorldRotation() {
		return new Quaternion(this.worldRotation);
	}

	/**
	 * Gets a direct reference to the inner <code>Quaternion</code> that
	 * contains scale in order to avoid memory allocations. Please, do
	 * <b>not</b> update rotation components by this way, use
	 * <code>rotateGlobally</code> for example.
	 * @return A reference to the rotation.
	 */
	public Quaternion getSharedWorldRotation() {
		return this.worldRotation;
	}

	public Vector3 getWorldTranslation(Vector3 result) {
		result.setValues(this.worldTranslation);
		return result;
	}

	public Vector3 getWorldTranslation() {
		return new Vector3(this.worldTranslation);
	}

	/**
	 * Gets a direct reference to the inner <code>Vector3</code> that contains
	 * translation in order to avoid memory allocations. Please, do <b>not</b>
	 * update translation components by this way, use <code>translateGlobally</code>
	 * for example.
	 * @return A reference to the translation.
	 */
	public Vector3 getSharedWorldTranslation() {
		return this.worldTranslation;
	}

	public Vector3 getLocalScale(Vector3 result) {
		result.setValues(this.localScale);
		return result;
	}

	public Vector3 getLocalScale() {
		return new Vector3(this.localScale);
	}

	/**
	 * Gets a direct reference to the inner <code>Vector3</code> that contains
	 * scale in order to avoid memory allocations. Please, do <b>not</b> update
	 * scale components by this way, use <code>scaleLocally</code> for example.
	 * @return A reference to the scale.
	 */
	public Vector3 getSharedLocalScale() {
		return this.localScale;
	}

	public Quaternion getLocalRotation(Quaternion result) {
		result.setValues(this.localRotation);
		return result;
	}

	public Quaternion getLocalRotation() {
		return new Quaternion(this.localRotation);
	}

	/**
	 * Gets a direct reference to the inner <code>Quaternion</code> that
	 * contains scale in order to avoid memory allocations. Please, do
	 * <b>not</b> update rotation components by this way, use
	 * <code>rotateLocally</code> for example.
	 * @return A reference to the rotation.
	 */
	public Quaternion getSharedLocalRotation() {
		return this.localRotation;
	}

	public Vector3 getLocalTranslation(Vector3 result) {
		result.setValues(this.localTranslation);
		return result;
	}

	public Vector3 getLocalTranslation() {
		return new Vector3(this.localTranslation);
	}

	/**
	 * Gets a direct reference to the inner <code>Vector3</code> that contains
	 * translation in order to avoid memory allocations. Please, do <b>not</b>
	 * update translation components by this way, use <code>translateLocally</code>
	 * for example.
	 * @return A reference to the translation.
	 */
	public Vector3 getSharedLocalTranslation() {
		return this.localTranslation;
	}

	public void rotateGlobally(Quaternion q) {
		this.worldRotation.mulSelf(q);
		this.worldDirty = true;
		this.inverseWorldDirty = true;
		updateLocalTransform();
	}

	public void translateGlobally(Vector3 translation) {
		translateGlobally(
				translation.getX(),
				translation.getY(),
				translation.getZ());
	}

	public void translateGlobally(float dx, float dy, float dz) {
		this.worldTranslation.addSelf(dx, dy, dz);
		this.worldDirty = true;
		this.inverseWorldDirty = true;
		updateLocalTransform();
	}

	public void scaleLocally(Vector3 scale) {
		scaleLocally(scale.getX(), scale.getY(), scale.getZ());
	}

	public void scaleLocally(float x, float y, float z) {
		this.localScale.scaleSelf(x, y, z);
		this.localDirty = true;
		this.inverseLocalDirty = true;
		updateWorldTransformAndChildren();
	}

	public void scaleLocally(float f) {
		scaleLocally(f, f, f);
	}

	public void rotateLocally(Quaternion q) {
		this.localRotation.mulSelf(q);
		this.localDirty = true;
		this.inverseLocalDirty = true;
		updateWorldTransformAndChildren();
	}

	public void translateLocally(Vector3 translation) {
		translateLocally(
				translation.getX(),
				translation.getY(),
				translation.getZ());
	}

	public void translateLocally(float dx, float dy, float dz) {
		this.localTranslation.addSelf(dx, dy, dz);
		this.localDirty = true;
		this.inverseLocalDirty = true;
		updateWorldTransformAndChildren();
	}

	/**
	 * Updates the world transform values from the local transform. This method
	 * should be called every time local coordinates have been modified.
	 * <br />
	 * The world transform is computed along the parent transform. If no parent
	 * has been provided, the world values will be the ones from the local
	 * transform.
	 * <br />
	 * In addition, the world transform values of the children will also
	 * be updated by calling this function on each of them.
	 */
	private void updateWorldTransformAndChildren() {
		if (hasParent()) {
			// s_world = s_world_parent * s_local
			this.worldDerivedScale.setValues(
					this.parent.worldDerivedScale.getX() * this.localScale.getX(),
					this.parent.worldDerivedScale.getY() * this.localScale.getY(),
					this.parent.worldDerivedScale.getZ() * this.localScale.getZ());
			// q_world = q_world_parent * q_local
			Quaternion.mul(this.worldRotation, this.parent.worldRotation, this.localRotation);
			// t_world = t_world_parent * t_local
			//  so with vectors => t_world_parent + t_local
			this.worldTranslation.setValues(
					this.parent.worldTranslation.getX() + this.localTranslation.getX(),
					this.parent.worldTranslation.getY() + this.localTranslation.getY(),
					this.parent.worldTranslation.getZ() + this.localTranslation.getZ());
		} else {
			this.worldDerivedScale.setValues(this.localScale);
			this.worldRotation.setValues(this.localRotation);
			this.worldTranslation.setValues(this.localTranslation);
		}
		// Updates children, if any.
		for (int i = 0, iEnd = this.children.size(); i < iEnd; i++) {
			this.children.get(i).updateWorldTransformAndChildren();
		}
		this.worldDirty = true;
		this.inverseWorldDirty = true;
	}

	/**
	 * Updates the local transform values from the global transform. This method
	 * should be called every time world coordinates have been modified.
	 * <br />
	 * The local transform is computed along the parent transform. If no parent
	 * has been provided, the local values will be the ones from the world
	 * transform.
	 */
	private void updateLocalTransform() {
		if (hasParent()) {
			// s_local = s_world_parent^-1 * s_world
			this.localScale.setValues(
					this.worldDerivedScale.getX() / this.parent.worldDerivedScale.getX(),
					this.worldDerivedScale.getY() / this.parent.worldDerivedScale.getY(),
					this.worldDerivedScale.getZ() / this.parent.worldDerivedScale.getZ());
			// q_local = q_world_parent^-1 * q_world
			Quaternion.inverse(this.localRotation, this.parent.worldRotation);
			this.localRotation.mulSelf(this.worldRotation);
			// t_local = t_world_parent^-1 * t_world
			//  so with vectors => t_world - t_world_parent
			this.localTranslation.setValues(
					this.worldTranslation.getX() - this.parent.worldTranslation.getX(),
					this.worldTranslation.getY() - this.parent.worldTranslation.getY(),
					this.worldTranslation.getZ() - this.parent.worldTranslation.getZ());
		} else {
			this.localScale.setValues(this.worldDerivedScale);
			this.localRotation.setValues(this.worldRotation);
			this.localTranslation.setValues(this.worldTranslation);
		}
		this.localDirty = true;
		this.inverseLocalDirty = true;
	}

	public Matrix4 getSharedWorldMatrix4() {
		updateWorldMatrixIfDirty();
		return this.worldMatrix;
	}

	public Matrix4 getSharedLocalMatrix4() {
		updateLocalMatrixIfDirty();
		return this.localMatrix;
	}

	public Matrix4 getSharedInverseWorldMatrix4() {
		updateCachedInverseWorldIfDirty();
		return this.cachedInverseWorldMatrix;
	}

	public Matrix4 getInverseWorldMatrix4(Matrix4 result) {
		updateCachedInverseWorldIfDirty();
		result.setValues(this.cachedInverseWorldMatrix);
		return result;
	}

	public Matrix4 getInverseWorldMatrix4() {
		final Matrix4 ret = new Matrix4();
		return getInverseWorldMatrix4(ret);
	}

	public float[] getInverseWorldMatrix4(float[] result) {
		updateCachedInverseWorldIfDirty();
		this.cachedInverseWorldMatrix.getComponents(result);
		return result;
	}

	public Matrix4 getSharedInverseLocalMatrix4() {
		updateCachedInverseLocalIfDirty();
		return this.cachedInverseLocalMatrix;
	}

	public Matrix4 getInverseLocalMatrix4(Matrix4 result) {
		updateCachedInverseLocalIfDirty();
		result.setValues(this.cachedInverseLocalMatrix);
		return result;
	}

	public Matrix4 getInverseLocalMatrix4() {
		final Matrix4 ret = new Matrix4();
		return getInverseLocalMatrix4(ret);
	}

	public float[] getInverseLocalMatrix4(float[] result) {
		updateCachedInverseLocalIfDirty();
		this.cachedInverseLocalMatrix.getComponents(result);
		return result;
	}

	private void updateCachedInverseWorldIfDirty() {
		updateWorldMatrixIfDirty();
		if (this.inverseWorldDirty) {
			this.worldMatrix.getInverse(this.cachedInverseWorldMatrix);
			this.inverseWorldDirty = false;
		}
	}

	private void updateCachedInverseLocalIfDirty() {
		updateLocalMatrixIfDirty();
		if (this.inverseLocalDirty) {
			this.localMatrix.getInverse(this.cachedInverseLocalMatrix);
			this.inverseLocalDirty = false;
		}
	}

	public static Matrix4 combineWorld(Matrix4 result, Transform lT, Transform rT) {
		lT.updateWorldMatrixIfDirty();
		// lT will be updated by the following combine.
		return combineWorld(result, lT, rT.worldMatrix);
	}

	public static float[] combineWorld(float[] result, Transform lT, Transform rT) {
		lT.updateWorldMatrixIfDirty();
		// lT will be updated by the following combine.
		return combineWorld(result, lT, rT.worldMatrix.getSharedComponents());
	}

	public static Matrix4 combineWorld(Matrix4 result, Transform lT, Matrix4 rM) {
		lT.updateWorldMatrixIfDirty();
		return Matrix4.combine(result, lT.worldMatrix, rM);
	}

	public static float[] combineWorld(float[] result, Transform lT, float[] rM) {
		lT.updateWorldMatrixIfDirty();
		Matrix.multiplyMM(result, 0, lT.worldMatrix.getSharedComponents(), 0, rM, 0);
		return result;
	}

	public static Matrix4 combineWorld(Transform lT, Matrix4 rM) {
		final Matrix4 ret = new Matrix4();
		return Matrix4.combine(ret, lT.worldMatrix, rM);
	}

	public static float[] combineWorld(Transform lT, float[] rM) {
		final float[] ret = new float[16];
		return combineWorld(ret, lT, rM);
	}

	public static Matrix4 combineWorld(Matrix4 result, Matrix4 lM, Transform rT) {
		rT.updateWorldMatrixIfDirty();
		return Matrix4.combine(result, lM, rT.worldMatrix);
	}

	public static float[] combineWorld(float[] result, float[] lM, Transform rT) {
		rT.updateWorldMatrixIfDirty();
		Matrix.multiplyMM(result, 0, lM, 0, rT.worldMatrix.getSharedComponents(), 0);
		return result;
	}

	public static float[] combineWorld(float[] lM, Transform rT) {
		final float[] ret = new float[16];
		return combineWorld(ret, lM, rT);
	}

	public static float[] combine(float[] result, float[] lM, float[] rM) {
		Matrix.multiplyMM(result, 0, lM, 0, rM, 0);
		return result;
	}

	public static float[] combine(float[] lM, float[] rM) {
		final float[] ret = new float[16];
		return combine(ret, lM, rM);
	}

	/**
	 * Gets a shared reference to the {@link Matrix4} that converts world to
	 * local coordinates:
	 * <pre>
	 *   local_coords = world_to_local * world_coords
	 * </pre>
	 * @return the world to local matrix.
	 */
	public Matrix4 getSharedWorldToLocalMatrix() {
		updateLocalMatrixIfDirty();
		return this.localMatrix;
	}

	/**
	 * Gets a shared reference to the {@link Matrix4} that converts local to
	 * world coordinates:
	 * <pre>
	 *   world_coords = local_to_world * local_coords
	 * </pre>
	 * @return the local to world matrix.
	 */
	public Matrix4 getSharedLocalToWorldMatrix() {
		updateCachedInverseLocalIfDirty();
		return this.cachedInverseLocalMatrix;
	}

	public Vector3 transformWorldDirection(Vector3 result, Vector3 v) {
		this.worldRotation.rotateDirection(result, v);
		return result;
	}

	public Vector3 transformWorldDirection(Vector3 v) {
		final Vector3 ret = new Vector3();
		return transformWorldDirection(ret, v);
	}

	public Vector3 transformWorldPoint(Vector3 result, Vector3 p) {
		updateWorldMatrixIfDirty();
		return this.worldMatrix.transformPoint(result, p);
	}

	public Vector3 transformWorldPoint(Vector3 v) {
		final Vector3 ret = new Vector3();
		return transformWorldPoint(ret, v);
	}

	public Vector3 transformLocalDirection(Vector3 result, Vector3 v) {
		this.localRotation.rotateDirection(result, v);
		return result;
	}

	public Vector3 transformLocalDirection(Vector3 v) {
		final Vector3 ret = new Vector3();
		return transformLocalDirection(ret, v);
	}

	public Vector3 transformLocalPoint(Vector3 result, Vector3 p) {
		updateLocalMatrixIfDirty();
		return this.localMatrix.transformPoint(result, p);
	}

	public Vector3 transformLocalPoint(Vector3 p) {
		final Vector3 ret = new Vector3();
		return transformLocalPoint(ret, p);
	}

	private static final Quaternion invLocalRotation = Quaternion.getNewIdQuat();

	public Vector3 transformWorldDirectionToLocal(Vector3 result, Vector3 v) {
		synchronized (invLocalRotation) {
			Quaternion.inverse(invLocalRotation, this.localRotation);
			invLocalRotation.rotateDirection(result, v);
		}
		return result;
	}

	public Vector3 transformWorldDirectionToLocal(Vector3 v) {
		final Vector3 ret = new Vector3();
		return transformWorldDirectionToLocal(ret, v);
	}

	public Vector3 transformWorldPointToLocal(Vector3 result, Vector3 v) {
		updateCachedInverseLocalIfDirty();
		return this.cachedInverseLocalMatrix.transformPoint(result, v);
	}

	public Vector3 transformWorldPointToLocal(Vector3 v) {
		final Vector3 ret = new Vector3();
		return transformWorldPointToLocal(ret, v);
	}
}
